# Copyright (c) 2023-2025, Songlin Yang, Yu Zhang

import warnings

import torch
import triton
import triton.language as tl

from fla.ops.attn.parallel import parallel_attn_bwd_preprocess
from fla.ops.nsa.compression import parallel_nsa_compression
from fla.ops.nsa.utils import _bitonic_merge
from fla.ops.utils import prepare_chunk_indices, prepare_chunk_offsets, prepare_lens, prepare_token_indices
from fla.ops.utils.op import exp, log
from fla.ops.utils.pooling import mean_pooling
from fla.utils import autocast_custom_bwd, autocast_custom_fwd, autotune_cache_kwargs, check_shared_mem, contiguous

try:
    from flash_attn import flash_attn_func, flash_attn_varlen_func
except ImportError:
    warnings.warn(
        "Flash Attention is not installed. Please install it via `pip install flash-attn --no-build-isolation`",
        category=ImportWarning,
    )
    flash_attn_func = None


@triton.heuristics({
    'IS_VARLEN': lambda args: args['cu_seqlens'] is not None,
})
@triton.autotune(
    configs=[
        triton.Config({}, num_warps=num_warps)
        for num_warps in [1, 2, 4]
    ],
    key=['BS', 'BK'],
    **autotune_cache_kwargs,
)
@triton.jit
def parallel_nsa_kernel_topk(
    q,
    k,
    lse,
    scale,
    block_indices,
    cu_seqlens,
    token_indices,
    chunk_offsets,
    T,
    H: tl.constexpr,
    HQ: tl.constexpr,
    G: tl.constexpr,
    K: tl.constexpr,
    S: tl.constexpr,
    BC: tl.constexpr,
    BS: tl.constexpr,
    BK: tl.constexpr,
    IS_VARLEN: tl.constexpr,
):
    i_t, i_bh = tl.program_id(0), tl.program_id(1)
    i_b, i_h = i_bh // H, i_bh % H

    if IS_VARLEN:
        i_n, i_t = tl.load(token_indices + i_t * 2).to(tl.int32), tl.load(token_indices + i_t * 2 + 1).to(tl.int32)
        bos, eos = tl.load(cu_seqlens + i_n).to(tl.int32), tl.load(cu_seqlens + i_n + 1).to(tl.int32)
        T = eos - bos
        boc = tl.load(chunk_offsets + i_n).to(tl.int32)
    else:
        bos, eos = i_b * T, i_b * T + T
        boc = i_b * tl.cdiv(T, BS)

    p_q = tl.make_block_ptr(q + (bos + i_t) * HQ*K, (HQ, K), (K, 1), (i_h * G, 0), (G, BK), (1, 0))

    # the Q block is kept in the shared memory throughout the whole kernel
    # [G, BK]
    b_q = tl.load(p_q, boundary_check=(0, 1))
    b_q = (b_q * scale).to(b_q.dtype)

    # the number of compression representations in total
    TC = tl.cdiv(T, BS)
    # the number of compression representations required to iterate over
    # incomplete compression blocks are not included
    NC = (i_t + 1) // BS
    ################################
    # 1. lse computation
    ################################
    if lse is not None:
        b_lse = tl.load(lse + (bos + i_t) * HQ + i_h * G + tl.arange(0, G))
    else:
        # max scores for the current block
        b_m = tl.full([G], float('-inf'), dtype=tl.float32)
        # lse = log(acc) + m
        b_acc = tl.zeros([G], dtype=tl.float32)
        for i_c in range(0, NC, BC):
            o_c = i_c + tl.arange(0, BC)

            p_k = tl.make_block_ptr(k + (boc * H + i_h) * K, (K, TC), (1, H*K), (0, i_c), (BK, BC), (0, 1))
            # [BK, BC]
            b_k = tl.load(p_k, boundary_check=(0, 1))

            # [G, BC]
            b_s = tl.dot(b_q, b_k)
            b_s = tl.where((o_c < NC)[None, :], b_s, float('-inf'))

            # [G]
            b_m, b_mp = tl.maximum(b_m, tl.max(b_s, 1)), b_m
            b_r = exp(b_mp - b_m)
            # [G, BC]
            b_p = exp(b_s - b_m[:, None])
            # [G]
            b_acc = b_acc * b_r + tl.sum(b_p, 1)

            b_mp = b_m
        if NC == 0:
            b_lse = tl.zeros([G], dtype=tl.float32)
        else:
            b_lse = b_m + log(b_acc)

    ################################
    # 2. topk selection
    ################################
    # [BC]
    b_i = tl.full([BC], -1, dtype=tl.float32)
    o_i = tl.zeros([BC], dtype=tl.int32)
    m_i = tl.arange(0, BC) < BC//2

    IC = i_t // BS
    for i_c in range(0, tl.cdiv(i_t + 1, BS), BC):
        o_c = i_c + tl.arange(0, BC)

        p_k = tl.make_block_ptr(k + (boc * H + i_h) * K, (K, TC), (1, H*K), (0, i_c), (BK, BC), (0, 1))
        # [BK, BC]
        b_k = tl.load(p_k, boundary_check=(0, 1))
        # [G, BC]
        b_s = tl.dot(b_q, b_k)
        b_s = tl.where(o_c < IC, b_s, float('-inf'))
        # [G, BC]
        # the 1st and the last 2 blocks are always selected
        b_p = tl.where((o_c == 0) | ((o_c == IC - 1) | (o_c == IC)), 1., exp(b_s - b_lse[:, None]))
        # the importance scores of the current block
        # [BC]
        b_i, b_ip = tl.sum(b_p, 0), b_i
        # blocks with index < 0 will be skipped
        o_i, o_ip = tl.where(o_c <= IC, o_c, -1), o_i

        n_dims: tl.constexpr = tl.standard._log2(b_i.shape[0])
        for i in tl.static_range(1, n_dims):
            b_i, o_i = _bitonic_merge(b_i, o_i.to(tl.int32), i, 2, n_dims)

        if i_c != 0:
            b_i, o_i = _bitonic_merge(b_i, o_i.to(tl.int32), n_dims, False, n_dims)
            b_i_new = b_ip * m_i + b_i * (1 - m_i)
            o_i_new = o_ip * m_i + o_i * (1 - m_i)
            b_i, o_i = _bitonic_merge(b_i_new, o_i_new.to(tl.int32), n_dims, True, n_dims)
        else:
            b_i, o_i = _bitonic_merge(b_i, o_i.to(tl.int32), n_dims, True, n_dims)

    m_top = tl.arange(0, BC//S) == 0
    b_top = tl.sum(m_top[:, None] * tl.reshape(o_i, [BC//S, S]), 0)

    p_b = tl.make_block_ptr(block_indices + (bos + i_t) * H*S, (H*S,), (1,), (i_h * S,), (S,), (0,))
    tl.store(p_b, b_top.to(p_b.dtype.element_ty))


@triton.heuristics({
    'IS_VARLEN': lambda args: args['cu_seqlens'] is not None,
    'USE_BLOCK_COUNTS': lambda args: isinstance(args['block_counts'], torch.Tensor),
})
@triton.autotune(
    configs=[
        triton.Config({}, num_warps=num_warps)
        for num_warps in [1, 2, 4]
    ],
    key=['BS', 'BK', 'BV'],
    **autotune_cache_kwargs,
)
@triton.jit
def parallel_nsa_fwd_kernel(
    q,
    k,
    v,
    o,
    lse,
    scale,
    block_indices,
    block_counts,
    cu_seqlens,
    token_indices,
    T,
    H: tl.constexpr,
    HQ: tl.constexpr,
    G: tl.constexpr,
    K: tl.constexpr,
    V: tl.constexpr,
    S: tl.constexpr,
    BS: tl.constexpr,
    BK: tl.constexpr,
    BV: tl.constexpr,
    IS_VARLEN: tl.constexpr,
    USE_BLOCK_COUNTS: tl.constexpr,
):
    i_t, i_v, i_bh = tl.program_id(0), tl.program_id(1), tl.program_id(2)
    i_b, i_h = i_bh // H, i_bh % H

    if IS_VARLEN:
        i_n, i_t = tl.load(token_indices + i_t * 2).to(tl.int32), tl.load(token_indices + i_t * 2 + 1).to(tl.int32)
        bos, eos = tl.load(cu_seqlens + i_n).to(tl.int32), tl.load(cu_seqlens + i_n + 1).to(tl.int32)
        T = eos - bos
    else:
        bos, eos = i_b * T, i_b * T + T

    k += (bos * H + i_h) * K
    v += (bos * H + i_h) * V
    block_indices += (bos + i_t) * H*S + i_h * S

    if USE_BLOCK_COUNTS:
        NS = tl.load(block_counts + (bos + i_t) * H + i_h)
    else:
        NS = S

    p_q = tl.make_block_ptr(q + (bos + i_t) * HQ*K, (HQ, K), (K, 1), (i_h * G, 0), (G, BK), (1, 0))
    # the Q block is kept in the shared memory throughout the whole kernel
    # [G, BK]
    b_q = tl.load(p_q, boundary_check=(0, 1))
    b_q = (b_q * scale).to(b_q.dtype)

    p_o = tl.make_block_ptr(o + (bos + i_t) * HQ*V, (HQ, V), (V, 1), (i_h * G, i_v * BV), (G, BV), (1, 0))
    p_lse = lse + (bos + i_t) * HQ + i_h * G + tl.arange(0, G)
    # [G, BV]
    b_o = tl.zeros([G, BV], dtype=tl.float32)

    b_m = tl.full([G], float('-inf'), dtype=tl.float32)
    b_acc = tl.zeros([G], dtype=tl.float32)
    for i in range(NS):
        i_s = tl.load(block_indices + i).to(tl.int32) * BS
        if i_s <= i_t and i_s >= 0:
            p_k = tl.make_block_ptr(k, (K, T), (1, H*K), (0, i_s), (BK, BS), (0, 1))
            p_v = tl.make_block_ptr(v, (T, V), (H*V, 1), (i_s, i_v * BV), (BS, BV), (1, 0))
            # [BK, BS]
            b_k = tl.load(p_k, boundary_check=(0, 1))
            # [BS, BV]
            b_v = tl.load(p_v, boundary_check=(0, 1))
            # [G, BS]
            b_s = tl.dot(b_q, b_k)
            b_s = tl.where((i_t >= (i_s + tl.arange(0, BS)))[None, :], b_s, float('-inf'))

            # [G]
            b_m, b_mp = tl.maximum(b_m, tl.max(b_s, 1)), b_m
            b_r = exp(b_mp - b_m)
            # [G, BS]
            b_p = exp(b_s - b_m[:, None])
            # [G]
            b_acc = b_acc * b_r + tl.sum(b_p, 1)
            # [G, BV]
            b_o = b_o * b_r[:, None] + tl.dot(b_p.to(b_q.dtype), b_v)

            b_mp = b_m
    b_o = b_o / b_acc[:, None]
    b_m += log(b_acc)
    tl.store(p_o, b_o.to(p_o.dtype.element_ty), boundary_check=(0, 1))
    tl.store(p_lse, b_m.to(p_lse.dtype.element_ty))


@triton.heuristics({
    'USE_BLOCK_COUNTS': lambda args: isinstance(args['block_counts'], torch.Tensor),
})
@triton.jit(do_not_specialize=['T'])
def parallel_nsa_kernel_mask(
    block_indices,
    block_counts,
    block_mask,
    T,
    H: tl.constexpr,
    S: tl.constexpr,
    BS: tl.constexpr,
    NS: tl.constexpr,
    USE_BLOCK_COUNTS: tl.constexpr,
):
    i_t, i_b, i_hs = tl.program_id(0), tl.program_id(1), tl.program_id(2)
    i_h, i_s = i_hs // S, i_hs % S

    b_i = tl.load(block_indices + i_b * T * H * S + i_t * H * S + i_h * S + i_s)
    if USE_BLOCK_COUNTS:
        b_m = b_i * BS <= i_t and i_s < tl.load(block_counts + i_b * T * H + i_t * H + i_h)
    else:
        b_m = b_i * BS <= i_t

    if b_i < NS and b_i >= 0:
        tl.store(block_mask + i_b * T * H * NS + i_t * H * NS + i_h * NS + b_i, b_m.to(block_mask.dtype.element_ty))


@triton.heuristics({
    'IS_VARLEN': lambda args: args['cu_seqlens'] is not None,
    'USE_BLOCK_COUNTS': lambda args: isinstance(args['block_counts'], torch.Tensor),
})
@triton.autotune(
    configs=[
        triton.Config({}, num_warps=num_warps)
        for num_warps in [1, 2, 4]
    ],
    key=['BS', 'BK', 'BV'],
    **autotune_cache_kwargs,
)
@triton.jit(do_not_specialize=['T'])
def parallel_nsa_bwd_kernel_dq(
    q,
    k,
    v,
    lse,
    delta,
    do,
    dq,
    scale,
    block_indices,
    block_counts,
    cu_seqlens,
    token_indices,
    T,
    B: tl.constexpr,
    H: tl.constexpr,
    HQ: tl.constexpr,
    G: tl.constexpr,
    K: tl.constexpr,
    V: tl.constexpr,
    S: tl.constexpr,
    BS: tl.constexpr,
    BK: tl.constexpr,
    BV: tl.constexpr,
    IS_VARLEN: tl.constexpr,
    USE_BLOCK_COUNTS: tl.constexpr,
):
    i_t, i_v, i_bh = tl.program_id(0), tl.program_id(1), tl.program_id(2)
    i_b, i_h = i_bh // H, i_bh % H

    all = B * T
    if IS_VARLEN:
        i_n, i_t = tl.load(token_indices + i_t * 2).to(tl.int32), tl.load(token_indices + i_t * 2 + 1).to(tl.int32)
        bos, eos = tl.load(cu_seqlens + i_n).to(tl.int32), tl.load(cu_seqlens + i_n + 1).to(tl.int32)
        T = eos - bos
    else:
        bos, eos = i_b * T, i_b * T + T

    q += (bos + i_t) * HQ*K
    do += (bos + i_t) * HQ*V
    lse += (bos + i_t) * HQ
    delta += (bos + i_t) * HQ
    dq += (i_v * all + bos + i_t) * HQ*K
    block_indices += (bos + i_t) * H*S + i_h * S

    if USE_BLOCK_COUNTS:
        NS = tl.load(block_counts + (bos + i_t) * H + i_h)
    else:
        NS = S

    k += (bos * H + i_h) * K
    v += (bos * H + i_h) * V

    p_q = tl.make_block_ptr(q, (HQ, K), (K, 1), (i_h * G, 0), (G, BK), (1, 0))
    p_dq = tl.make_block_ptr(dq, (HQ, K), (K, 1), (i_h * G, 0), (G, BK), (1, 0))

    # [G, BK]
    b_q = tl.load(p_q, boundary_check=(0, 1))
    b_q = (b_q * scale).to(b_q.dtype)

    p_do = tl.make_block_ptr(do, (HQ, V), (V, 1), (i_h * G, i_v * BV), (G, BV), (1, 0))
    p_lse = lse + i_h * G + tl.arange(0, G)
    p_delta = delta + i_h * G + tl.arange(0, G)

    # [G, BV]
    b_do = tl.load(p_do, boundary_check=(0, 1))
    # [G]
    b_lse = tl.load(p_lse)
    b_delta = tl.load(p_delta)

    # [G, BK]
    b_dq = tl.zeros([G, BK], dtype=tl.float32)
    for i in range(NS):
        i_s = tl.load(block_indices + i).to(tl.int32) * BS
        if i_s <= i_t and i_s >= 0:
            p_k = tl.make_block_ptr(k, (K, T), (1, H*K), (0, i_s), (BK, BS), (0, 1))
            p_v = tl.make_block_ptr(v, (V, T), (1, H*V), (i_v * BV, i_s), (BV, BS), (0, 1))
            # [BK, BS]
            b_k = tl.load(p_k, boundary_check=(0, 1))
            # [BV, BS]
            b_v = tl.load(p_v, boundary_check=(0, 1))

            # [G, BS]
            b_s = tl.dot(b_q, b_k)
            b_p = exp(b_s - b_lse[:, None])
            b_p = tl.where((i_t >= (i_s + tl.arange(0, BS)))[None, :], b_p, 0)

            # [G, BV] @ [BV, BS] -> [G, BS]
            b_dp = tl.dot(b_do, b_v)
            b_ds = b_p * (b_dp.to(tl.float32) - b_delta[:, None])
            # [G, BS] @ [BS, BK] -> [G, BK]
            b_dq += tl.dot(b_ds.to(b_k.dtype), tl.trans(b_k))
    b_dq *= scale

    tl.store(p_dq, b_dq.to(p_dq.dtype.element_ty), boundary_check=(0, 1))


@triton.heuristics({
    'IS_VARLEN': lambda args: args['cu_seqlens'] is not None,
})
@triton.autotune(
    configs=[
        triton.Config({}, num_warps=num_warps)
        for num_warps in [1, 2, 4]
    ],
    key=['BS', 'BK', 'BV'],
    **autotune_cache_kwargs,
)
@triton.jit(do_not_specialize=['T'])
def parallel_nsa_bwd_kernel_dkv(
    q,
    k,
    v,
    lse,
    delta,
    do,
    dk,
    dv,
    block_mask,
    cu_seqlens,
    chunk_indices,
    scale,
    T,
    B: tl.constexpr,
    H: tl.constexpr,
    HQ: tl.constexpr,
    G: tl.constexpr,
    K: tl.constexpr,
    V: tl.constexpr,
    M: tl.constexpr,
    BS: tl.constexpr,
    BK: tl.constexpr,
    BV: tl.constexpr,
    IS_VARLEN: tl.constexpr,
):
    i_v, i_s, i_bh = tl.program_id(0), tl.program_id(1), tl.program_id(2)
    i_b, i_h = i_bh // H, i_bh % H

    all = B * T
    if IS_VARLEN:
        i_n, i_s = tl.load(chunk_indices + i_s * 2).to(tl.int32), tl.load(chunk_indices + i_s * 2 + 1).to(tl.int32)
        bos, eos = tl.load(cu_seqlens + i_n).to(tl.int32), tl.load(cu_seqlens + i_n + 1).to(tl.int32)
        T = eos - bos
    else:
        bos, eos = i_b * T, i_b * T + T

    p_k = tl.make_block_ptr(k + (bos * H + i_h) * K, (T, K), (H*K, 1), (i_s * BS, 0), (BS, BK), (1, 0))
    p_v = tl.make_block_ptr(v + (bos * H + i_h) * V, (T, V), (H*V, 1), (i_s * BS, i_v * BV), (BS, BV), (1, 0))
    p_dk = tl.make_block_ptr(dk + (i_v * all * H + bos * H + i_h) * K, (T, K), (H*K, 1), (i_s * BS, 0), (BS, BK), (1, 0))
    p_dv = tl.make_block_ptr(dv + (bos * H + i_h) * V, (T, V), (H*V, 1), (i_s * BS, i_v * BV), (BS, BV), (1, 0))

    # [BS, BK]
    b_k = tl.load(p_k, boundary_check=(0, 1))
    b_dk = tl.zeros([BS, BK], dtype=tl.float32)
    # [BS, BV]
    b_v = tl.load(p_v, boundary_check=(0, 1))
    b_dv = tl.zeros([BS, BV], dtype=tl.float32)

    for i in range(i_s * BS, T):
        b_m = tl.load(block_mask + (bos + i) * H*M + i_h * M + i_s)
        if b_m:
            p_q = tl.make_block_ptr(q + (bos + i) * HQ*K, (HQ, K), (K, 1), (i_h * G, 0), (G, BK), (1, 0))
            # [G, BK]
            b_q = tl.load(p_q, boundary_check=(0, 1))
            b_q = (b_q * scale).to(b_q.dtype)

            p_do = tl.make_block_ptr(do + (bos + i) * HQ*V, (HQ, V), (V, 1), (i_h * G, i_v * BV), (G, BV), (1, 0))
            p_lse = lse + (bos + i) * HQ + i_h * G + tl.arange(0, G)
            p_delta = delta + (bos + i) * HQ + i_h * G + tl.arange(0, G)
            # [G, BV]
            b_do = tl.load(p_do, boundary_check=(0, 1))
            # [G]
            b_lse = tl.load(p_lse)
            b_delta = tl.load(p_delta)
            # [BS, G]
            b_s = tl.dot(b_k, tl.trans(b_q))
            b_p = exp(b_s - b_lse[None, :])
            b_p = tl.where((i >= (i_s * BS + tl.arange(0, BS)))[:, None], b_p, 0)
            # [BS, G] @ [G, BV] -> [BS, BV]
            b_dv += tl.dot(b_p.to(b_do.dtype), b_do)
            # [BS, BV] @ [BV, G] -> [BS, G]
            b_dp = tl.dot(b_v, tl.trans(b_do))
            # [BS, G]
            b_ds = b_p * (b_dp - b_delta[None, :])
            # [BS, G] @ [G, BK] -> [BS, BK]
            b_dk += tl.dot(b_ds.to(b_q.dtype), b_q)

    tl.store(p_dk, b_dk.to(p_dk.dtype.element_ty), boundary_check=(0, 1))
    tl.store(p_dv, b_dv.to(p_dv.dtype.element_ty), boundary_check=(0, 1))


def parallel_nsa_topk(
    q: torch.Tensor,
    k: torch.Tensor,
    lse: torch.Tensor,
    block_counts: torch.LongTensor | int,
    block_size: int = 64,
    scale: float = None,
    cu_seqlens: torch.LongTensor | None = None,
) -> torch.LongTensor:
    B, T, HQ, K = q.shape
    H = k.shape[2]
    G = HQ // H
    # the number of selected blocks for each token
    S = block_counts if isinstance(block_counts, int) else block_counts.max().item()
    S = triton.next_power_of_2(S)
    # here we set BC = BS, but beware that they can be chosen separately if required
    BC = BS = block_size
    BK = max(triton.next_power_of_2(K), 16)
    assert BC >= 2 * S, f"BC ({BC}) must be greater than or equal to 2 * S ({S})"

    block_indices = torch.zeros(B, T, H, S, dtype=torch.int32, device=q.device)
    token_indices = prepare_token_indices(cu_seqlens) if cu_seqlens is not None else None
    chunk_offsets = prepare_chunk_offsets(cu_seqlens, BS) if cu_seqlens is not None else None
    grid = (T, B * H)
    # the 1st and the last 2 blocks are always selected
    parallel_nsa_kernel_topk[grid](
        q=q,
        k=k,
        lse=lse,
        scale=scale,
        block_indices=block_indices,
        cu_seqlens=cu_seqlens,
        token_indices=token_indices,
        chunk_offsets=chunk_offsets,
        T=T,
        H=H,
        HQ=HQ,
        G=G,
        K=K,
        S=S,
        BC=BC,
        BS=BS,
        BK=BK,
    )
    return block_indices


def parallel_nsa_fwd(
    q: torch.Tensor,
    k: torch.Tensor,
    v: torch.Tensor,
    block_indices: torch.LongTensor,
    block_counts: torch.LongTensor | int,
    block_size: int,
    scale: float,
    cu_seqlens: torch.LongTensor | None = None,
    token_indices: torch.LongTensor | None = None,
):
    B, T, H, K, V, S = *k.shape, v.shape[-1], block_indices.shape[-1]
    HQ = q.shape[2]
    G = HQ // H
    BS = block_size
    if check_shared_mem('hopper', q.device.index):
        BK = min(256, triton.next_power_of_2(K))
        BV = min(256, triton.next_power_of_2(V))
    else:
        BK = min(128, triton.next_power_of_2(K))
        BV = min(128, triton.next_power_of_2(V))
    NK = triton.cdiv(K, BK)
    NV = triton.cdiv(V, BV)
    assert NK == 1, "The key dimension can not be larger than 256"

    grid = (T, NV, B * H)
    o = torch.empty(B, T, HQ, V, dtype=v.dtype, device=q.device)
    lse = torch.empty(B, T, HQ, dtype=torch.float, device=q.device)

    parallel_nsa_fwd_kernel[grid](
        q=q,
        k=k,
        v=v,
        o=o,
        lse=lse,
        scale=scale,
        block_indices=block_indices,
        block_counts=block_counts,
        cu_seqlens=cu_seqlens,
        token_indices=token_indices,
        T=T,
        H=H,
        HQ=HQ,
        G=G,
        K=K,
        V=V,
        S=S,
        BS=BS,
        BK=BK,
        BV=BV,
    )
    return o, lse


def parallel_nsa_block_mask(
    block_indices: torch.LongTensor,
    block_counts: torch.LongTensor | int,
    cu_seqlens: torch.LongTensor,
    block_size: int,
):
    B, T, H, S = block_indices.shape
    BS = block_size
    if cu_seqlens is not None:
        NS = triton.cdiv(prepare_lens(cu_seqlens).max().item(), BS)
    else:
        NS = triton.cdiv(T, BS)
    block_mask = torch.zeros(B, T, H, NS, dtype=torch.bool, device=block_indices.device)

    parallel_nsa_kernel_mask[(T, B, H*S)](
        block_indices=block_indices,
        block_counts=block_counts,
        block_mask=block_mask,
        T=T,
        H=H,
        S=S,
        BS=BS,
        NS=NS,
    )
    return block_mask


def parallel_nsa_bwd(
    q: torch.Tensor,
    k: torch.Tensor,
    v: torch.Tensor,
    o: torch.Tensor,
    lse: torch.Tensor,
    do: torch.Tensor,
    block_indices: torch.Tensor,
    block_counts: torch.LongTensor | int,
    block_size: int = 64,
    scale: float = None,
    cu_seqlens: torch.LongTensor | None = None,
    token_indices: torch.LongTensor | None = None,
):
    B, T, H, K, V, S = *k.shape, v.shape[-1], block_indices.shape[-1]
    HQ = q.shape[2]
    G = HQ // H
    BS = block_size
    BK = max(triton.next_power_of_2(K), 16)
    BV = min(128, max(triton.next_power_of_2(v.shape[-1]), 16))
    NV = triton.cdiv(V, BV)

    delta = parallel_attn_bwd_preprocess(o, do)

    dq = torch.empty(NV, *q.shape, dtype=q.dtype if NV == 1 else torch.float, device=q.device)
    grid = (T, NV, B * H)
    parallel_nsa_bwd_kernel_dq[grid](
        q=q,
        k=k,
        v=v,
        lse=lse,
        delta=delta,
        do=do,
        dq=dq,
        block_indices=block_indices,
        block_counts=block_counts,
        cu_seqlens=cu_seqlens,
        token_indices=token_indices,
        scale=scale,
        T=T,
        B=B,
        H=H,
        HQ=HQ,
        G=G,
        K=K,
        V=V,
        S=S,
        BS=BS,
        BK=BK,
        BV=BV,
    )
    dq = dq.sum(0)

    if cu_seqlens is not None:
        chunk_indices = prepare_chunk_indices(cu_seqlens, BS)
        NS = len(chunk_indices)
    else:
        chunk_indices = None
        NS = triton.cdiv(T, BS)

    # [B, T, H, M]
    block_mask = parallel_nsa_block_mask(block_indices, block_counts, cu_seqlens, block_size)
    dk = torch.empty(NV, *k.shape, dtype=k.dtype if NV == 1 else torch.float, device=q.device)
    dv = torch.empty(v.shape, dtype=v.dtype, device=q.device)

    grid = (NV, NS, B * H)
    parallel_nsa_bwd_kernel_dkv[grid](
        q=q,
        k=k,
        v=v,
        lse=lse,
        delta=delta,
        do=do,
        dk=dk,
        dv=dv,
        block_mask=block_mask,
        cu_seqlens=cu_seqlens,
        chunk_indices=chunk_indices,
        scale=scale,
        T=T,
        B=B,
        H=H,
        HQ=HQ,
        G=G,
        K=K,
        V=V,
        M=block_mask.shape[-1],
        BS=BS,
        BK=BK,
        BV=BV,
    )
    dk = dk.sum(0)
    return dq, dk, dv


@torch.compile
class ParallelNSAFunction(torch.autograd.Function):

    @staticmethod
    @contiguous
    @autocast_custom_fwd
    def forward(ctx, q, k, v, block_indices, block_counts, block_size, scale, cu_seqlens):
        ctx.dtype = q.dtype

        # 2-d sequence indices denoting the cu_seqlens of tokens in each sequence
        # for example, if the passed `cu_seqlens` is [0, 2, 6],
        # then there are 2 and 4 tokens in the 1st and 2nd sequences respectively, and `token_indices` will be
        # [[0, 0], [0, 1], [1, 0], [1, 1], [1, 2], [1, 3]]
        token_indices = prepare_token_indices(cu_seqlens) if cu_seqlens is not None else None

        o, lse = parallel_nsa_fwd(
            q=q,
            k=k,
            v=v,
            block_indices=block_indices,
            block_counts=block_counts,
            block_size=block_size,
            scale=scale,
            cu_seqlens=cu_seqlens,
            token_indices=token_indices,
        )
        ctx.save_for_backward(q, k, v, o, lse)
        ctx.block_indices = block_indices
        ctx.block_counts = block_counts
        ctx.cu_seqlens = cu_seqlens
        ctx.token_indices = token_indices
        ctx.block_size = block_size
        ctx.scale = scale
        return o.to(q.dtype)

    @staticmethod
    @contiguous
    @autocast_custom_bwd
    def backward(ctx, do):
        q, k, v, o, lse = ctx.saved_tensors
        dq, dk, dv = parallel_nsa_bwd(
            q=q,
            k=k,
            v=v,
            o=o,
            lse=lse,
            do=do,
            block_indices=ctx.block_indices,
            block_counts=ctx.block_counts,
            block_size=ctx.block_size,
            scale=ctx.scale,
            cu_seqlens=ctx.cu_seqlens,
            token_indices=ctx.token_indices,
        )
        return dq.to(q), dk.to(k), dv.to(v), None, None, None, None, None, None, None, None


def parallel_nsa(
    q: torch.Tensor,
    k: torch.Tensor,
    v: torch.Tensor,
    g_cmp: torch.Tensor | None = None,
    g_slc: torch.Tensor | None = None,
    g_swa: torch.Tensor | None = None,
    block_indices: torch.LongTensor | None = None,
    block_counts: torch.LongTensor | int = 16,
    block_size: int = 64,
    window_size: int = 0,
    scale: float | None = None,
    cu_seqlens: torch.LongTensor | None = None,
) -> torch.Tensor:
    r"""
    Args:
        q (torch.Tensor):
            queries of shape `[B, T, HQ, K]`.
        k (torch.Tensor):
            keys of shape `[B, T, H, K]`.
            GQA is enforced here. The ratio of query heads (HQ) to key/value heads (H) must be a power of 2 and >=16.
        v (torch.Tensor):
            values of shape `[B, T, H, V]`.
        g_cmp (torch.Tensor):
            Gate score for compressed attention of shape `[B, T, HQ]`.
        g_slc (torch.Tensor):
            Gate score for selected attention of shape `[B, T, HQ]`.
        g_swa (torch.Tensor):
            Gate score for sliding attentionof shape `[B, T, HQ]`.
        block_indices (torch.LongTensor):
            Block indices of shape `[B, T, H, S]`.
            `S` is the number of selected blocks for each query token, which is set to 16 in the paper.
            If `g_cmp` is provided, the passed `block_indices` will be ignored.
        block_counts (Optional[Union[torch.LongTensor, int]]):
            Number of selected blocks for each query.
            If a tensor is provided, with shape `[B, T, H]`,
            each query can select the same number of blocks.
            If not provided, it will default to 16.
        block_size (int):
            Selected block size. Default: 64.
        window_size (int):
            Sliding window size. Default: 0.
        scale (Optional[float]):
            Scale factor for attention scores.
            If not provided, it will default to `1 / sqrt(K)`. Default: `None`.
        cu_seqlens (torch.LongTensor):
            Cumulative sequence lengths of shape `[N+1]` used for variable-length training,
            consistent with the FlashAttention API.

    Returns:
        o (torch.Tensor):
            Outputs of shape `[B, T, HQ, V]`.
    """
    assert block_counts is not None, "block counts must be provided for selection"
    if scale is None:
        scale = k.shape[-1] ** -0.5
    if cu_seqlens is not None:
        assert q.shape[0] == 1, "batch size must be 1 when cu_seqlens are provided"
    assert q.shape[2] % (k.shape[2] * 16) == 0, "Group size must be a multiple of 16 in NSA"

    k_cmp, v_cmp = mean_pooling(k, block_size, cu_seqlens), mean_pooling(v, block_size, cu_seqlens)
    o_cmp, lse_cmp = None, None
    if g_cmp is not None:
        o_cmp, lse_cmp = parallel_nsa_compression(
            q=q,
            k=k_cmp,
            v=v_cmp,
            block_size=block_size,
            scale=scale,
            cu_seqlens=cu_seqlens,
        )
        if block_indices is not None:
            warnings.warn("`block_indices` will be ignored when `g_cmp` is provided")
        block_indices = parallel_nsa_topk(
            q=q,
            k=k_cmp,
            lse=lse_cmp,
            block_counts=block_counts,
            block_size=block_size,
            scale=scale,
            cu_seqlens=cu_seqlens,
        )
    o = o_slc = ParallelNSAFunction.apply(q, k, v, block_indices, block_counts, block_size, scale, cu_seqlens)
    if g_slc is not None:
        o = o_slc * g_slc.unsqueeze(-1)
    if o_cmp is not None:
        o = torch.addcmul(o, o_cmp, g_cmp.unsqueeze(-1))
    if window_size > 0:
        if cu_seqlens is not None:
            max_seqlen = q.shape[1]
            o_swa = flash_attn_varlen_func(
                q.squeeze(0), k.squeeze(0), v.squeeze(0),
                cu_seqlens_q=cu_seqlens,
                cu_seqlens_k=cu_seqlens,
                max_seqlen_q=max_seqlen,
                max_seqlen_k=max_seqlen,
                causal=True,
                window_size=(window_size-1, 0),
            ).unsqueeze(0)
        else:
            o_swa = flash_attn_func(
                q, k, v,
                causal=True,
                window_size=(window_size-1, 0),
            )
        o = torch.addcmul(o, o_swa, g_swa.unsqueeze(-1))
    return o



"""Quantization primitives implemented in triton."""


import triton
import triton.language as tl


@triton.jit
def clamp(x: tl.tensor, min_val, max_val) -> tl.tensor:
    """Clamps all elements in `x` into range [min, max].

    Args:
        x (tl.tensor): the input tensor.
        min_val (Number): lower bound of the range.
        max_val (Number): upper bound of the range.

    Returns:
        tl.tensor: the output tensor.
    """
    return tl.math.min(tl.math.max(x, min_val), max_val)


@triton.jit
def dequantize(x: tl.tensor, scale: tl.tensor) -> tl.tensor:
    """Dequantize quantized tensor to floating point.

    Args:
        x (tl.tensor): quantized tensor.
        scale (tl.tensor): quantization scaling factor

    Returns:
        tl.tensor: Dequantized floating-point tensor.
    """
    return (x * scale).to(tl.float32)


@triton.jit
def quantize(x, scale, qmin, qmax) -> tl.tensor:
    """Quantize the tensor given quantization scale and data type.

    Args:
        x (tl.tensor): floating-point tensor
        scale (tl.tensor): quantization scale factor.
        qmin (Number): quantization minimum range.
        qmax (Number): quantization maximum range

    Returns:
        tl.tensor: rounded and clamped tensor.
            Note: this is still in floating point as we can't pass dtype to function

    Example:
    
        out = quantize(out, scale, -128, 127).to(tl.int8)
    """
    return clamp(tl.math.round(x / scale), qmin, qmax)



"""
Resize kernel - Pure Triton
ATen: aten::upsample_nearest2d, aten::upsample_bilinear2d
"""
import torch
import triton
import triton.language as tl
from neurobrix.kernels.registry import register_kernel


@triton.jit
def _resize_nearest_kernel(
    x_ptr, out_ptr,
    N, C, H_in, W_in,
    H_out, W_out,
    scale_h, scale_w,
    stride_xn, stride_xc, stride_xh, stride_xw,
    stride_on, stride_oc, stride_oh, stride_ow,
    BLOCK_SIZE: tl.constexpr
):
    """Nearest neighbor resize - pure Triton."""
    pid = tl.program_id(0)
    n_idx = tl.program_id(1)
    c_idx = tl.program_id(2)

    # Each block handles BLOCK_SIZE output pixels
    hw_start = pid * BLOCK_SIZE
    hw_offs = hw_start + tl.arange(0, BLOCK_SIZE)
    mask = hw_offs < H_out * W_out

    h_out = hw_offs // W_out
    w_out = hw_offs % W_out

    # Map to input coordinates (nearest)
    h_in = (h_out.to(tl.float32) / scale_h).to(tl.int32)
    w_in = (w_out.to(tl.float32) / scale_w).to(tl.int32)

    # Clamp
    h_in = tl.minimum(tl.maximum(h_in, 0), H_in - 1)
    w_in = tl.minimum(tl.maximum(w_in, 0), W_in - 1)

    # Load from input
    x_offs = n_idx * stride_xn + c_idx * stride_xc + h_in * stride_xh + w_in * stride_xw
    val = tl.load(x_ptr + x_offs, mask=mask, other=0.0)

    # Store to output
    out_offs = n_idx * stride_on + c_idx * stride_oc + h_out * stride_oh + w_out * stride_ow
    tl.store(out_ptr + out_offs, val, mask=mask)


@triton.jit
def _resize_bilinear_kernel(
    x_ptr, out_ptr,
    N, C, H_in, W_in,
    H_out, W_out,
    scale_h, scale_w,
    stride_xn, stride_xc, stride_xh, stride_xw,
    stride_on, stride_oc, stride_oh, stride_ow,
    BLOCK_SIZE: tl.constexpr
):
    """Bilinear resize - pure Triton."""
    pid = tl.program_id(0)
    n_idx = tl.program_id(1)
    c_idx = tl.program_id(2)

    hw_start = pid * BLOCK_SIZE
    hw_offs = hw_start + tl.arange(0, BLOCK_SIZE)
    mask = hw_offs < H_out * W_out

    h_out = hw_offs // W_out
    w_out = hw_offs % W_out

    # Map to input coordinates
    h_in_f = (h_out.to(tl.float32) + 0.5) / scale_h - 0.5
    w_in_f = (w_out.to(tl.float32) + 0.5) / scale_w - 0.5

    # Get integer and fractional parts
    h0 = tl.floor(h_in_f).to(tl.int32)
    w0 = tl.floor(w_in_f).to(tl.int32)
    h1 = h0 + 1
    w1 = w0 + 1

    # Fractional parts
    fh = h_in_f - h0.to(tl.float32)
    fw = w_in_f - w0.to(tl.float32)

    # Clamp coordinates
    h0 = tl.minimum(tl.maximum(h0, 0), H_in - 1)
    h1 = tl.minimum(tl.maximum(h1, 0), H_in - 1)
    w0 = tl.minimum(tl.maximum(w0, 0), W_in - 1)
    w1 = tl.minimum(tl.maximum(w1, 0), W_in - 1)

    # Load 4 corners
    base = n_idx * stride_xn + c_idx * stride_xc
    v00 = tl.load(x_ptr + base + h0 * stride_xh + w0 * stride_xw, mask=mask, other=0.0)
    v01 = tl.load(x_ptr + base + h0 * stride_xh + w1 * stride_xw, mask=mask, other=0.0)
    v10 = tl.load(x_ptr + base + h1 * stride_xh + w0 * stride_xw, mask=mask, other=0.0)
    v11 = tl.load(x_ptr + base + h1 * stride_xh + w1 * stride_xw, mask=mask, other=0.0)

    # Bilinear interpolation
    val = (v00 * (1.0 - fh) * (1.0 - fw) +
           v01 * (1.0 - fh) * fw +
           v10 * fh * (1.0 - fw) +
           v11 * fh * fw)

    # Store
    out_offs = n_idx * stride_on + c_idx * stride_oc + h_out * stride_oh + w_out * stride_ow
    tl.store(out_ptr + out_offs, val, mask=mask)


def _resize_impl(x: torch.Tensor, size: list, mode: str = "nearest") -> torch.Tensor:
    """Resize implementation using Triton."""
    assert x.is_cuda
    x = x.contiguous()

    # Assume NCHW format
    N, C, H_in, W_in = x.shape
    H_out, W_out = size

    scale_h = float(H_out) / float(H_in)
    scale_w = float(W_out) / float(W_in)

    out = torch.empty(N, C, H_out, W_out, dtype=x.dtype, device=x.device)

    BLOCK_SIZE = 256
    num_hw = H_out * W_out
    grid = (triton.cdiv(num_hw, BLOCK_SIZE), N, C)

    with torch.cuda.device(x.device):
        if mode == "nearest":
            _resize_nearest_kernel[grid](
                x, out,
                N, C, H_in, W_in,
                H_out, W_out,
                scale_h, scale_w,
                x.stride(0), x.stride(1), x.stride(2), x.stride(3),
                out.stride(0), out.stride(1), out.stride(2), out.stride(3),
                BLOCK_SIZE=BLOCK_SIZE,
                num_warps=4,
            )
        else:  # bilinear
            _resize_bilinear_kernel[grid](
                x, out,
                N, C, H_in, W_in,
                H_out, W_out,
                scale_h, scale_w,
                x.stride(0), x.stride(1), x.stride(2), x.stride(3),
                out.stride(0), out.stride(1), out.stride(2), out.stride(3),
                BLOCK_SIZE=BLOCK_SIZE,
                num_warps=4,
            )

    return out


@register_kernel(family="llm", vendor="nvidia", tier="common", op_name="resize")
def resize_kernel(x: torch.Tensor, size: list = None, mode: str = "nearest") -> torch.Tensor:
    """Resize tensor - pure Triton."""
    if size is None:
        return x
    return _resize_impl(x, size, mode)




import logging
import math

import torch
import triton
import triton.language as tl

from flag_gems import runtime
from flag_gems.runtime import torch_device_fn
from flag_gems.utils import libentry

from ..utils import TOTAL_CORE_NUM, cfggen_reduce_op2, count_divisible_by_2

logger = logging.getLogger("flag_gems").getChild(__name__.lstrip("."))


@triton.jit
def reduce_mul(a, b):
    return a * b


@libentry()
@triton.autotune(configs=cfggen_reduce_op2(), key=["M"])
@triton.jit
def prod_kernel_mid(
    inp,
    mid,
    M,
    BLOCK_SIZE: tl.constexpr,
    ITER_NUM: tl.constexpr,
):
    pid = tl.program_id(0)
    num_jobs = tl.num_programs(axis=0)
    block_start = pid * BLOCK_SIZE
    step = num_jobs * BLOCK_SIZE
    _tmp = tl.full([BLOCK_SIZE], value=1.0, dtype=tl.float32)
    block_start = block_start.to(tl.int64)
    for off in range(block_start, M, step):
        offset = off + tl.arange(0, BLOCK_SIZE)
        mask = offset < M
        inp_val = tl.load(inp + offset, mask=mask, other=1.0).to(tl.float32)
        _tmp = inp_val * _tmp

    # Reset to original reduce programming mode after optimizing the tl.reduce.
    for x in tl.static_range(1, int(ITER_NUM), 1):
        _tmp[: BLOCK_SIZE // (2**x)] = (
            _tmp[: BLOCK_SIZE // (2**x)]
            * _tmp[BLOCK_SIZE // (2**x) : (BLOCK_SIZE // (2**x)) * 2]
        )

    mid_ptr = mid + pid
    tl.store(mid_ptr, _tmp[0])


@libentry()
@triton.jit
def prod_kernel_result(mid, out, mid_size: tl.constexpr, loop_num: tl.constexpr):
    offset = tl.arange(0, mid_size)
    mid_val = tl.load(mid + offset)

    # Reset to original reduce programming mode after optimizing the tl.reduce.
    for x in tl.static_range(1, loop_num, 1):
        mid_val[: mid_size // (2**x)] = (
            mid_val[: mid_size // (2**x)]
            * mid_val[mid_size // (2**x) : (mid_size // (2**x)) * 2]
        )

    prod_val = tl.reduce(
        mid_val[: mid_size // (2 ** (loop_num - 1))], axis=0, combine_fn=reduce_mul
    )
    tl.store(out, prod_val)


def prod(inp, *, dtype=None):
    logger.debug("GEMS_CAMBRICON PROD")
    if dtype is None:
        dtype = inp.dtype

    M = inp.numel()
    grid = lambda meta: (min(triton.cdiv(M, meta["BLOCK_SIZE"]), TOTAL_CORE_NUM),)
    mid_size = TOTAL_CORE_NUM
    loop_num = count_divisible_by_2(mid_size) + 1

    mid = torch.ones((mid_size,), dtype=dtype, device=inp.device)
    out = torch.empty([], dtype=dtype, device=inp.device)

    with torch_device_fn.device(inp.device):
        prod_kernel_mid[grid](inp, mid, M)
        prod_kernel_result[(1, 1, 1)](mid, out, mid_size, loop_num)
    return out


def heur_block_n(args):
    return triton.next_power_of_2(args["N"])


@libentry()
@triton.autotune(
    configs=runtime.get_tuned_config("prod"),
    key=[
        "M",
        "N",
    ],
)
@triton.jit
def prod_kernel(
    inp,
    out,
    M,
    N,
    K,
    BLOCK_M: tl.constexpr,
    BLOCK_N: tl.constexpr,
):
    # set offset
    pid_m = tl.program_id(0)
    pid_k = tl.program_id(1)
    m_offset = pid_m * BLOCK_M + tl.arange(0, BLOCK_M)

    acc = tl.full((BLOCK_M, BLOCK_N), value=1.0, dtype=tl.float32)
    for start_n in range(0, N, BLOCK_N):
        n_offset = start_n + tl.arange(0, BLOCK_N)
        offset = m_offset[:, None] * N * K + n_offset[None, :] * K + pid_k

        # set mask
        mask = m_offset[:, None] < M and n_offset[None, :] < N
        inp_ptrs = inp + offset
        inp_vals = tl.load(inp_ptrs, mask=mask, other=1.0).to(tl.float32)
        acc *= inp_vals
    result_index = tl.reduce(acc, axis=1, combine_fn=reduce_mul)

    offset_index = m_offset * K + pid_k
    out_ptrs = out + offset_index
    mask1 = m_offset < M
    tl.store(out_ptrs, result_index, mask=mask1)


def prod_dim(inp, dim=None, keepdim=False, *, dtype=None):
    logger.debug("GEMS_CAMBRICON PROD DIM")

    assert dim >= -inp.ndim and dim < inp.ndim, "Invalid dim"
    shape = inp.shape
    dim = dim % inp.ndim
    N = shape[dim]
    M = math.prod(shape[:dim])
    K = inp.numel() // M // N

    inp = inp.contiguous()

    shape_list = list(shape)
    shape_list[dim] = 1

    if dtype is None:
        dtype = inp.dtype
    out = torch.empty(shape_list, dtype=dtype, device=inp.device)
    if not keepdim:
        out = torch.squeeze(out, dim)

    grid = lambda meta: (
        triton.cdiv(M, meta["BLOCK_M"]),
        K,
    )
    with torch_device_fn.device(inp.device):
        prod_kernel[grid](inp, out, M, N, K)

    return out



"""
DualAR Engine — Fish-Speech/OpenAudio DualAR Flow

Handles DualAR architecture: backbone generates semantic tokens from
3D input [B, N+1, T], then embeds tokens for codec decoder → waveform.

ZERO SEMANTIC: No knowledge of "Fish-Speech" or "OpenAudio".
ZERO HARDCODE: All parameters from NBX container.
"""

import gc
import time
import torch
from typing import Any, Callable, Dict, List, Optional

from .base import FlowHandler, FlowContext, register_flow


@register_flow("dual_ar")
class DualAREngine(FlowHandler):
    """
    DualAR semantic token generation + codec decoding.

    topology.flow.audio:
        direction: tts
        stages:
          - component: model
            execution: dual_ar
          - component: codec.decoder
            execution: forward
    """

    def __init__(
        self,
        ctx: FlowContext,
        execute_component_fn: Callable,
        resolve_inputs_fn: Callable,
        ensure_weights_fn: Callable,
        unload_weights_fn: Callable,
    ):
        super().__init__(ctx)
        self._execute_component = execute_component_fn
        self._resolve_component_inputs = resolve_inputs_fn
        self._ensure_weights_loaded = ensure_weights_fn
        self._unload_component_weights = unload_weights_fn

    def execute(self) -> Dict[str, Any]:
        """Execute DualAR pipeline."""
        flow = self.ctx.pkg.topology.get("flow", {})
        audio_config = flow.get("audio", {})
        stages = audio_config.get("stages", [])
        defaults = self.ctx.pkg.defaults

        if not stages:
            raise RuntimeError("ZERO FALLBACK: dual_ar flow requires at least one stage.")

        # ── Step 1: Tokenize text input ──
        from .audio_utils import preprocess_text_input
        preprocess_text_input(self.ctx)

        # ── Step 2: DualAR generation ──
        backbone_stage = stages[0]
        comp_name = backbone_stage["component"]
        device = self.ctx.primary_device

        max_tokens = defaults.get("max_tokens", 2048)
        temperature = defaults.get("temperature", 0.7)
        top_p = defaults.get("top_p", 0.8)
        eos_token_id = defaults.get("eos_token_id")

        print(f"   [{comp_name}] DualAR generation (max_tokens={max_tokens})...")
        start = time.perf_counter()

        self._ensure_weights_loaded(comp_name)
        executor = self.ctx.executors[comp_name]

        # Get graph input shape: [B, N+1, T]
        dag = getattr(executor, '_dag', None)
        n_codebooks = 11
        trace_seq_len = 64
        if dag:
            for _tid, spec in dag.get("tensors", {}).items():
                if spec.get("input_name") == "inp":
                    shape = spec.get("shape", [])
                    if len(shape) >= 3:
                        n_val = shape[1]
                        n_codebooks = n_val if isinstance(n_val, int) else 11
                        t_val = shape[2]
                        trace_seq_len = t_val if isinstance(t_val, int) else 64
                    break

        # Get tokenized input
        input_ids = self.ctx.variable_resolver.resolved.get("global.input_ids")
        if input_ids is None:
            raise RuntimeError("ZERO FALLBACK: DualAR requires tokenized input_ids.")

        # Build initial 3D input [B, N+1, T_prompt]
        prompt_len = input_ids.shape[-1]
        inp = torch.zeros(1, n_codebooks, prompt_len, dtype=torch.long, device=device)
        inp[0, 0, :prompt_len] = input_ids.squeeze(0)[:prompt_len]

        # Generate semantic tokens autoregressively
        generated_semantic: List[int] = []

        for step in range(max_tokens):
            # Pad/truncate to trace_seq_len for CompiledSequence
            cur_len = prompt_len + len(generated_semantic)
            if cur_len <= trace_seq_len:
                padded = torch.zeros(1, n_codebooks, trace_seq_len, dtype=torch.long, device=device)
                padded[0, 0, :prompt_len] = input_ids.squeeze(0)[:prompt_len]
                for i, tok in enumerate(generated_semantic):
                    padded[0, 0, prompt_len + i] = tok
            else:
                # Sliding window
                all_semantic = list(input_ids.squeeze(0).tolist()[:prompt_len]) + generated_semantic
                window = all_semantic[-trace_seq_len:]
                padded = torch.zeros(1, n_codebooks, trace_seq_len, dtype=torch.long, device=device)
                for i, tok in enumerate(window):
                    padded[0, 0, i] = tok

            output = executor.run({"inp": padded})

            if isinstance(output, dict):
                logits = next(iter(output.values()))
            elif isinstance(output, torch.Tensor):
                logits = output
            else:
                logits = None

            if logits is None:
                break

            # Get logits at last real token position
            pos = min(cur_len - 1, trace_seq_len - 1)
            if logits.dim() == 3:
                last_logits = logits[:, pos, :].clone()
            else:
                last_logits = logits.clone()

            from .audio_utils import sample_token
            next_token = sample_token(
                last_logits.unsqueeze(1), temperature, top_p=top_p,
            )

            if eos_token_id is not None and next_token == eos_token_id:
                break

            generated_semantic.append(next_token)

        elapsed = (time.perf_counter() - start) * 1000
        print(f"   [{comp_name}] Generated {len(generated_semantic)} semantic tokens in {elapsed:.0f}ms")

        # ── Step 3: Embed semantic tokens → codec input ──
        # Find the main text embedding (largest vocab, excludes codebook/fast)
        embed_weight = None
        if hasattr(executor, '_weights'):
            best_vocab = 0
            for wname, wtensor in executor._weights.items():
                if 'embed' in wname and wname.endswith('.weight'):
                    if 'codebook' in wname or 'fast' in wname:
                        continue
                    if wtensor.shape[0] > best_vocab:
                        best_vocab = wtensor.shape[0]
                        embed_weight = wtensor

        if embed_weight is None:
            raise RuntimeError("ZERO FALLBACK: DualAR model must have embed.weight.")

        token_ids = torch.tensor(generated_semantic, dtype=torch.long, device=device)
        max_id = token_ids.max().item()
        print(f"   [{comp_name}] Token range: 0..{max_id}, embed vocab: {embed_weight.shape[0]}")
        if max_id >= embed_weight.shape[0]:
            print(f"   [{comp_name}] WARNING: token {max_id} >= vocab {embed_weight.shape[0]}, clamping")
            token_ids = token_ids.clamp(0, embed_weight.shape[0] - 1)
        with torch.no_grad():
            token_embeds = torch.nn.functional.embedding(token_ids, embed_weight)

        codec_input = token_embeds.unsqueeze(0).transpose(1, 2)  # [1, dim, T_gen]
        print(f"   [{comp_name}] Embedded → {list(codec_input.shape)} for codec.decoder")

        # Store for downstream
        self.ctx.variable_resolver.resolved[f"{comp_name}.output_0"] = codec_input
        self.ctx.variable_resolver.resolved["global.generated_codes"] = token_ids
        self.ctx.variable_resolver.resolved["global.generated_token_ids"] = generated_semantic

        if not self.ctx.persistent_mode:
            self._unload_component_weights(comp_name)
            gc.collect()
            device_empty_cache(self.ctx.primary_device)

        # ── Step 4: Codec decoder (forward stages) ──
        for stage in stages[1:]:
            codec_name = stage["component"]
            if codec_name not in self.ctx.executors:
                print(f"   [{codec_name}] Skipped (not in executors)")
                continue

            print(f"   [{codec_name}] Running forward pass...")
            codec_start = time.perf_counter()
            self._ensure_weights_loaded(codec_name)

            # Try chunked forward if input exceeds trace-time seq_len
            if not self._try_chunked_forward(codec_name):
                self._execute_component(codec_name, "forward", None)

            codec_elapsed = (time.perf_counter() - codec_start) * 1000
            print(f"   [{codec_name}] Done in {codec_elapsed:.0f}ms")

            if not self.ctx.persistent_mode:
                self._unload_component_weights(codec_name)
                gc.collect()
                device_empty_cache(self.ctx.primary_device)

        # ── Step 5: Output waveform ──
        from .audio_utils import postprocess_audio_output
        postprocess_audio_output(self.ctx)

        return self.ctx.variable_resolver.resolve_all()

    def _try_chunked_forward(self, comp_name: str) -> bool:
        """Run chunked forward if input seq_len exceeds graph's trace-time seq_len."""
        executor = self.ctx.executors.get(comp_name)
        if executor is None:
            return False

        dag = getattr(executor, '_dag', None)
        if dag is None:
            return False

        graph_input_name = None
        graph_seq_len = None
        for _tid, spec in dag.get("tensors", {}).items():
            iname = spec.get("input_name")
            if iname:
                shape = spec.get("shape", [])
                if len(shape) == 3:
                    graph_seq_len = shape[2]
                    if isinstance(graph_seq_len, dict):
                        graph_seq_len = graph_seq_len.get("trace_value", graph_seq_len)
                    graph_input_name = iname
                break

        if graph_input_name is None or not isinstance(graph_seq_len, int):
            return False

        # Find actual input tensor
        resolved = self.ctx.variable_resolver.resolved
        actual_input = None
        for key in [f"global.{graph_input_name}", graph_input_name]:
            val = resolved.get(key)
            if isinstance(val, torch.Tensor) and val.dim() == 3:
                actual_input = val
                break

        if actual_input is None:
            connections = self.ctx.pkg.topology.get("connections", [])
            for conn in connections:
                if conn.get("to", "") == f"{comp_name}.{graph_input_name}":
                    val = resolved.get(conn.get("from", ""))
                    if isinstance(val, torch.Tensor) and val.dim() == 3:
                        actual_input = val
                        break

        if actual_input is None:
            return False

        actual_seq = actual_input.shape[2]
        if actual_seq <= graph_seq_len:
            return False

        print(f"   [{comp_name}] Chunked: {actual_seq} frames → {graph_seq_len}-frame blocks"
              f" (actual_input shape={list(actual_input.shape)})")
        waveform_chunks = []

        for chunk_start in range(0, actual_seq, graph_seq_len):
            chunk_end = min(chunk_start + graph_seq_len, actual_seq)
            chunk = actual_input[:, :, chunk_start:chunk_end]

            if chunk.shape[2] < graph_seq_len:
                pad_size = graph_seq_len - chunk.shape[2]
                chunk = torch.nn.functional.pad(chunk, (0, pad_size))

            print(f"     chunk [{chunk_start}:{chunk_end}] shape={list(chunk.shape)} dtype={chunk.dtype}")
            output = executor.run({graph_input_name: chunk})

            if isinstance(output, dict):
                out_tensor = next(iter(output.values()))
            elif isinstance(output, torch.Tensor):
                out_tensor = output
            else:
                out_tensor = None

            if out_tensor is not None:
                if chunk_end - chunk_start < graph_seq_len and out_tensor.dim() >= 2:
                    ratio = (chunk_end - chunk_start) / graph_seq_len
                    trim_len = int(out_tensor.shape[-1] * ratio)
                    out_tensor = out_tensor[..., :trim_len]
                waveform_chunks.append(out_tensor)

        if waveform_chunks:
            full_output = torch.cat(waveform_chunks, dim=-1)
            self.ctx.variable_resolver.resolved[f"{comp_name}.output_0"] = full_output
            self.ctx.variable_resolver.resolved["global.output_audio"] = full_output
            print(f"   [{comp_name}] Waveform: {list(full_output.shape)}")

        return True



"""
NeuroBrix Validators - Base Classes.

Defines the foundation for all validation in NeuroBrix:
- SafetyLevel enum for validation intensity
- Base exceptions with rich context
- ValidationResult for structured reporting
- BaseValidator abstract class

ZERO FALLBACK: All validation failures raise explicit exceptions.
"""

from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from enum import IntEnum
from typing import Any, Dict, List, Optional, TYPE_CHECKING

if TYPE_CHECKING:
    pass  # Future: from ..graph import GraphNode


class SafetyLevel(IntEnum):
    """
    Validation intensity levels.

    MINIMAL: Fast checks only (production hot path)
    STANDARD: Balanced safety/performance (default)
    PARANOID: All checks enabled (debugging/development)

    Uses IntEnum to support comparison operators (< > <= >=).
    """
    MINIMAL = 1
    STANDARD = 2
    PARANOID = 3


# ============================================================================
# Exceptions
# ============================================================================

class ValidationError(Exception):
    """
    Base exception for all validation failures.

    All validation exceptions carry rich context for debugging.
    """

    def __init__(
        self,
        message: str,
        *,
        node: Optional[Any] = None,
        context: Optional[Dict[str, Any]] = None,
        suggestion: Optional[str] = None,
    ):
        self.node = node
        self.context = context or {}
        self.suggestion = suggestion

        # Build rich error message
        parts = [message]
        if node is not None:
            node_name = getattr(node, 'name', str(node))
            node_op = getattr(node, 'op', 'unknown')
            parts.append(f"Node: {node_name} (op={node_op})")
        if context:
            parts.append(f"Context: {context}")
        if suggestion:
            parts.append(f"Suggestion: {suggestion}")

        super().__init__("\n".join(parts))


class ShapeViolation(ValidationError):
    """Raised when tensor shapes are incompatible."""

    def __init__(
        self,
        message: str,
        *,
        expected_shape: Optional[tuple] = None,
        actual_shape: Optional[tuple] = None,
        **kwargs
    ):
        self.expected_shape = expected_shape
        self.actual_shape = actual_shape

        context = kwargs.pop('context', {})
        if expected_shape is not None:
            context['expected_shape'] = expected_shape
        if actual_shape is not None:
            context['actual_shape'] = actual_shape

        super().__init__(message, context=context, **kwargs)


class MemoryPlanError(ValidationError):
    """Raised when memory plan is invalid or exceeds limits."""

    def __init__(
        self,
        message: str,
        *,
        required_bytes: Optional[int] = None,
        available_bytes: Optional[int] = None,
        **kwargs
    ):
        self.required_bytes = required_bytes
        self.available_bytes = available_bytes

        context = kwargs.pop('context', {})
        if required_bytes is not None:
            context['required_bytes'] = required_bytes
            context['required_gb'] = required_bytes / (1024**3)
        if available_bytes is not None:
            context['available_bytes'] = available_bytes
            context['available_gb'] = available_bytes / (1024**3)

        super().__init__(message, context=context, **kwargs)


class NumericalDivergence(ValidationError):
    """Raised when numerical instability is detected."""

    def __init__(
        self,
        message: str,
        *,
        nan_count: Optional[int] = None,
        inf_count: Optional[int] = None,
        max_value: Optional[float] = None,
        **kwargs
    ):
        self.nan_count = nan_count
        self.inf_count = inf_count
        self.max_value = max_value

        context = kwargs.pop('context', {})
        if nan_count is not None:
            context['nan_count'] = nan_count
        if inf_count is not None:
            context['inf_count'] = inf_count
        if max_value is not None:
            context['max_value'] = max_value

        super().__init__(message, context=context, **kwargs)


class UnsupportedOpsError(ValidationError):
    """Raised when graph contains unsupported operations."""

    def __init__(
        self,
        message: str,
        *,
        unsupported_ops: Optional[List[str]] = None,
        **kwargs
    ):
        self.unsupported_ops = unsupported_ops or []

        context = kwargs.pop('context', {})
        if unsupported_ops:
            context['unsupported_ops'] = unsupported_ops

        super().__init__(message, context=context, **kwargs)


class SyncError(ValidationError):
    """Raised when synchronization issues are detected."""

    def __init__(
        self,
        message: str,
        *,
        device_a: Optional[str] = None,
        device_b: Optional[str] = None,
        **kwargs
    ):
        self.device_a = device_a
        self.device_b = device_b

        context = kwargs.pop('context', {})
        if device_a is not None:
            context['device_a'] = device_a
        if device_b is not None:
            context['device_b'] = device_b

        super().__init__(message, context=context, **kwargs)


class FusionError(ValidationError):
    """Raised when kernel fusion is invalid."""

    def __init__(
        self,
        message: str,
        *,
        nodes_to_fuse: Optional[List[str]] = None,
        reason: Optional[str] = None,
        **kwargs
    ):
        self.nodes_to_fuse = nodes_to_fuse or []
        self.reason = reason

        context = kwargs.pop('context', {})
        if nodes_to_fuse:
            context['nodes_to_fuse'] = nodes_to_fuse
        if reason:
            context['fusion_failure_reason'] = reason

        super().__init__(message, context=context, **kwargs)


class NaNInfDetected(NumericalDivergence):
    """Specific exception for NaN/Inf detection."""
    pass


class DeadlockRisk(SyncError):
    """Specific exception for potential deadlock situations."""
    pass


# ============================================================================
# Validation Result
# ============================================================================

@dataclass
class ValidationResult:
    """
    Structured result from validation.

    Contains pass/fail status plus detailed diagnostics.
    """
    passed: bool
    validator_name: str
    message: str = ""
    warnings: List[str] = field(default_factory=list)
    metrics: Dict[str, Any] = field(default_factory=dict)

    def __bool__(self) -> bool:
        return self.passed

    def raise_if_failed(self) -> None:
        """Raise ValidationError if validation failed."""
        if not self.passed:
            raise ValidationError(
                f"Validation failed: {self.validator_name}\n{self.message}"
            )


# ============================================================================
# Base Validator
# ============================================================================

class BaseValidator(ABC):
    """
    Abstract base class for all validators.

    Validators check specific aspects of the execution graph:
    - Shapes compatibility
    - Memory feasibility
    - Numerical stability
    - Operation support
    - Synchronization correctness
    - Fusion validity
    """

    def __init__(self, safety_level: SafetyLevel = SafetyLevel.STANDARD):
        self.safety_level = safety_level

    @property
    @abstractmethod
    def name(self) -> str:
        """Validator name for reporting."""
        pass

    @abstractmethod
    def validate(self, graph: Any, context: Dict[str, Any]) -> ValidationResult:
        """
        Validate the graph.

        Args:
            graph: The execution graph to validate
            context: Additional context (memory limits, device info, etc.)

        Returns:
            ValidationResult with pass/fail and diagnostics

        Note:
            Implementations should NOT raise exceptions for validation failures.
            Instead, return ValidationResult(passed=False, ...).
            Exceptions should only be raised for internal errors.
        """
        pass

    def should_run(self, safety_level: SafetyLevel) -> bool:
        """
        Check if this validator should run at the given safety level.

        Override in subclasses to customize when validator runs.
        Default: run at all levels.
        """
        return True



import triton
import triton.language as tl
import torch 

@triton.jit()
def _a100_quantized_matmul(a_ptr, b_ptr, c_ptr, scales_ptr, zeros_ptr,
                             stride_am, stride_ak,
                             stride_bk, stride_bn,
                             stride_cm, stride_cn,
                             stride_scales_g, stride_scales_n,
                             stride_zeros_g, stride_zeros_n,
                             groupsize,
                             m, n, k,
                             block_size_m: tl.constexpr, block_size_n: tl.constexpr, block_size_k: tl.constexpr,
                             group_size_m: tl.constexpr,
                             ):
    
    pid = tl.program_id(0)

    total_blocks_m = tl.cdiv(m, block_size_m)
    total_blocks_n = tl.cdiv(n, block_size_n)
    total_blocks_k = tl.cdiv(k, block_size_k)

    num_blocks_in_group = group_size_m * total_blocks_n
    group_id = pid // num_blocks_in_group
    group_size = min(total_blocks_m - group_id * group_size_m, group_size_m)

    pid_m = group_id * group_size_m + (pid % group_size)
    pid_n = (pid % num_blocks_in_group) // (group_size)

    offs_m = (pid_m * block_size_m + tl.arange(0, block_size_m)) % m
    offs_n = (pid_n * block_size_n + tl.arange(0, block_size_n)) % n

    offs_am = tl.max_contiguous(tl.multiple_of(offs_m, block_size_m), block_size_m)
    offs_bn = tl.max_contiguous(tl.multiple_of(offs_n, block_size_n), block_size_n)
    offs_k = tl.arange(0, block_size_k)
    
    a_ptrs = a_ptr + (offs_am[:, None] * stride_am + offs_k[None, :] * stride_ak)
    b_ptrs = b_ptr + ((offs_k[:, None] // 8) * stride_bk + offs_bn[None, :] * stride_bn)
    
    scales_ptrs = scales_ptr + offs_bn * stride_scales_n
    zeros_ptrs = zeros_ptr + ((offs_bn // 8) * stride_zeros_n)

    shifter = (offs_k % 8) * 4
    zeros_shifter = (offs_bn % 8) * 4


    output = tl.zeros((block_size_m, block_size_n), dtype=tl.float32)
    for k in range(0, total_blocks_k):
 
        a = tl.load(a_ptrs)
        b = tl.load(b_ptrs)
        g_id = k // (groupsize // block_size_k)

        ptr = scales_ptrs + g_id * stride_scales_g
        scales = tl.load(ptr)
        
        ptr = zeros_ptrs + g_id * stride_zeros_g
        zeros = tl.load(ptr)

        zeros = (zeros >> zeros_shifter) & 0xF
        zeros = (zeros + 1) * scales

        b = (b >> shifter[:, None]) & 0xF # b -> int32
        b = b * scales[None, :] - zeros[None, :] # b -> fp16
        
        output += tl.dot(a, b)
        a_ptrs += stride_ak * block_size_k
        b_ptrs +=  (block_size_k//8) * stride_bk
    
    output.to(tl.float16)
    offs_cm = pid_m * block_size_m + tl.arange(0, block_size_m)
    offs_cn = pid_n * block_size_n + tl.arange(0, block_size_n)
    c_ptrs = c_ptr + (offs_cm[:, None] * stride_cm + offs_cn[None, :] * stride_cn)
    tl.store(c_ptrs, output)

class a100_qlinear(torch.autograd.Function):
    def forward(ctx, a, b, scales, zeros):

        m, k = a.shape
        _, n = b.shape

        quant_groupsize = 128
        block_size_m = 16 
        block_size_n = 32 # [N = 4096 // 32] = 128 blocks
        block_size_k = 256
        group_size_m = 8
        num_warps = 4
        num_stages = 8
        total_blocks_m = triton.cdiv(m, block_size_m)
        total_blocks_n = triton.cdiv(n, block_size_n)
        total_programs  = total_blocks_m * total_blocks_n
        grid = (total_programs, 1)

        c = torch.zeros((m, n), device=b.device, dtype=torch.float16)
        k = _a100_quantized_matmul[grid](
            a, b, c, scales, zeros,
            a.stride(0), a.stride(1),
            b.stride(0), b.stride(1),
            c.stride(0), c.stride(1),
            scales.stride(0), scales.stride(1),
            zeros.stride(0), zeros.stride(1),
            quant_groupsize,
            m, n, k,
            block_size_m, block_size_n, block_size_k, group_size_m,
            num_warps = num_warps, num_stages = num_stages,
        )

        print(f"{k.n_regs} registers used, {k.n_spills} spills, {k.shared/1000} kB shared memory\n")

        with open('dequant_simple.txt', 'w') as f:

            print(f"{k.n_regs} registers used, {k.n_spills} spills, {k.shared/1000} kB shared memory\n", file=f)
            print("IR", k.asm['ttir'], file=f)
            print("TTGIR", k.asm['ttgir'], file=f)
            print("PTX", k.asm['ptx'], file=f)
            print(f"{k.n_regs} registers used, {k.n_spills} spills, {k.shared/1000} kB shared memory\n", file=f)

            print(f"{total_blocks_m=} x {total_blocks_n=} = {total_programs=}")
        return c
        

a100_qlinear = a100_qlinear.apply


"""
neurobrix chat — Interactive multi-turn chat with running daemon.

Connects to ServingDaemon via Unix socket. Supports:
    /new     — Start new conversation
    /context — Show token usage
    /status  — Show engine status
    /quit    — Exit chat

Usage:
    neurobrix chat
    neurobrix chat --temperature 0.7
"""

import sys


def cmd_chat(args):
    """Interactive chat REPL connected to serving daemon."""
    from neurobrix.serving.client import DaemonClient

    # Check daemon is running
    if not DaemonClient.is_running():
        print("[Chat] No daemon running.")
        print("[Chat] Start one first:")
        print("  neurobrix serve --model <name>")
        sys.exit(1)

    # Connect
    client = DaemonClient()
    try:
        client.connect()
    except RuntimeError as e:
        print(f"[Chat] Connection failed: {e}")
        sys.exit(1)

    # Get status for display
    try:
        status = client.status()
    except RuntimeError as e:
        print(f"[Chat] Failed to get status: {e}")
        client.close()
        sys.exit(1)

    model_name = status.get("model", "unknown")
    family = status.get("family", "unknown")

    if family != "llm":
        print(f"[Chat] Model family is '{family}', not 'llm'.")
        print(f"[Chat] Use 'neurobrix run' for non-LLM models.")
        client.close()
        sys.exit(1)

    # Build generation kwargs from CLI args
    gen_kwargs = {}
    if getattr(args, 'max_tokens', None) is not None:
        gen_kwargs["max_tokens"] = args.max_tokens
    if getattr(args, 'temperature', None) is not None:
        gen_kwargs["temperature"] = args.temperature
    if getattr(args, 'repetition_penalty', None) is not None:
        gen_kwargs["repetition_penalty"] = args.repetition_penalty

    print(f"\nNeuroBrix Chat — {model_name}")
    print(f"Commands: /new /context /status /quit")
    print(f"{'─' * 50}\n")

    try:
        while True:
            try:
                user_input = input("You: ").strip()
            except EOFError:
                break

            if not user_input:
                continue

            # Slash commands
            if user_input.startswith("/"):
                cmd = user_input.lower().split()[0]

                if cmd in ("/quit", "/exit", "/q"):
                    break

                elif cmd == "/new":
                    client.new_chat()
                    print("[New conversation started]\n")
                    continue

                elif cmd == "/context":
                    try:
                        st = client.status()
                        ctx = st.get("context", {})
                        if ctx:
                            turns = ctx.get("turns", 0)
                            tokens = ctx.get("estimated_tokens", 0)
                            max_ctx = ctx.get("max_context", 0)
                            usage = ctx.get("usage_pct", 0)
                            print(f"[Context] {turns} messages, ~{tokens} tokens, "
                                  f"{usage}% of {max_ctx} max\n")
                        else:
                            print("[Context] No active conversation\n")
                    except RuntimeError as e:
                        print(f"[Error] {e}\n")
                    continue

                elif cmd == "/status":
                    try:
                        st = client.status()
                        print(f"[Status] Model: {st.get('model')}")
                        print(f"[Status] Family: {st.get('family')}")
                        print(f"[Status] Mode: {st.get('mode')}")
                        print(f"[Status] VRAM: {st.get('vram_used_gb', '?')} GB\n")
                    except RuntimeError as e:
                        print(f"[Error] {e}\n")
                    continue

                else:
                    print(f"[Unknown command: {cmd}]")
                    print(f"Commands: /new /context /status /quit\n")
                    continue

            # Send chat message
            try:
                result = client.chat(user_input, **gen_kwargs)
                if result.get("summarized"):
                    print(f"\n[Context compressed — older turns summarized]")
                text = result.get("text", "")
                if text:
                    print(f"\nAssistant: {text}\n")
                else:
                    print(f"\n[No response generated]\n")
            except RuntimeError as e:
                print(f"\n[Error] {e}\n")

    except KeyboardInterrupt:
        print("\n")

    finally:
        client.close()
        print("[Chat] Disconnected")



"""
Universal Div - Pure Triton (Robust)
"""
import torch
import triton
import triton.language as tl

@triton.jit
def _div_kernel(x_ptr, y_ptr, out_ptr, n_elements, BLOCK_SIZE: tl.constexpr):
    pid = tl.program_id(axis=0)
    block_start = pid * BLOCK_SIZE
    offsets = block_start + tl.arange(0, BLOCK_SIZE)
    mask = offsets < n_elements

    x = tl.load(x_ptr + offsets, mask=mask)
    y = tl.load(y_ptr + offsets, mask=mask)

    # Avoid div by zero - use a safer threshold
    y_safe = tl.where(tl.abs(y) < 1e-7, tl.where(y >= 0, 1e-7, -1e-7), y)

    output = x / y_safe
    # Clamp output to prevent fp16 overflow
    output = tl.minimum(tl.maximum(output, -65000.0), 65000.0)
    tl.store(out_ptr + offsets, output, mask=mask)



# Copyright (c) 2023-2024, Yu Zhang, Songlin Yang

from typing import Optional, Tuple

import torch
import triton
import triton.language as tl

from ....torch.utils import contiguous
from ..fla_utils import chunk_reversed_cumsum_fwd


@triton.autotune(
    configs=[
        triton.Config({"BS": 16}, num_warps=2),
        triton.Config({"BS": 16}, num_warps=4),
        triton.Config({"BS": 16}, num_warps=8),
        triton.Config({"BS": 32}, num_warps=2),
        triton.Config({"BS": 32}, num_warps=4),
        triton.Config({"BS": 32}, num_warps=8),
        triton.Config({"BS": 64}, num_warps=2),
        triton.Config({"BS": 64}, num_warps=4),
        triton.Config({"BS": 64}, num_warps=8),
    ],
    key=["S"],
)
@triton.jit
def chunk_gla_fwd_kernel_cum(
    s,
    o,
    s_s_h,
    s_s_t,
    s_s_d,
    T: tl.constexpr,
    S: tl.constexpr,
    BT: tl.constexpr,
    BS: tl.constexpr,
):
    i_s, i_t, i_bh = tl.program_id(0), tl.program_id(1), tl.program_id(2)
    o_i = tl.arange(0, BT)
    m_s = tl.where(o_i[:, None] >= o_i[None, :], 1.0, 0.0)

    p_s = tl.make_block_ptr(
        s + i_bh * s_s_h, (T, S), (s_s_t, s_s_d), (i_t * BT, i_s * BS), (BT, BS), (1, 0)
    )
    p_o = tl.make_block_ptr(
        o + i_bh * s_s_h, (T, S), (s_s_t, s_s_d), (i_t * BT, i_s * BS), (BT, BS), (1, 0)
    )
    # [BT, BS]
    b_s = tl.load(p_s, boundary_check=(0, 1)).to(tl.float32)
    b_o = tl.dot(m_s, b_s, allow_tf32=False)
    tl.store(p_o, b_o.to(p_o.dtype.element_ty), boundary_check=(0, 1))


@triton.jit
def chunk_gla_fwd_kernel_h(
    k,
    v,
    g,
    h,
    h0,
    ht,
    s_k_h,
    s_k_t,
    s_k_d,
    s_v_h,
    s_v_t,
    s_v_d,
    s_h_h,
    s_h_t,
    s_h_d,
    T: tl.constexpr,
    K: tl.constexpr,
    V: tl.constexpr,
    BT: tl.constexpr,
    BK: tl.constexpr,
    BV: tl.constexpr,
    NT: tl.constexpr,
    USE_INITIAL_STATE: tl.constexpr,
    STORE_FINAL_STATE: tl.constexpr,
):
    i_v, i_k, i_bh = tl.program_id(0), tl.program_id(1), tl.program_id(2)
    b_h = tl.zeros([BK, BV], dtype=tl.float32)
    if USE_INITIAL_STATE:
        p_h = tl.make_block_ptr(
            h0 + i_bh * K * V, (K, V), (V, 1), (i_k * BK, i_v * BV), (BK, BV), (1, 0)
        )
        b_h += tl.load(p_h, boundary_check=(0, 1)).to(tl.float32)
    for i_t in range(NT):
        p_k = tl.make_block_ptr(
            k + i_bh * s_k_h,
            (K, T),
            (s_k_d, s_k_t),
            (i_k * BK, i_t * BT),
            (BK, BT),
            (0, 1),
        )
        p_v = tl.make_block_ptr(
            v + i_bh * s_v_h,
            (T, V),
            (s_v_t, s_v_d),
            (i_t * BT, i_v * BV),
            (BT, BV),
            (1, 0),
        )
        p_h = tl.make_block_ptr(
            h + i_bh * s_h_h + i_t * K * V,
            (K, V),
            (s_h_t, s_h_d),
            (i_k * BK, i_v * BV),
            (BK, BV),
            (1, 0),
        )
        p_g = tl.make_block_ptr(
            g + i_bh * s_k_h,
            (K, T),
            (s_k_d, s_k_t),
            (i_k * BK, i_t * BT),
            (BK, BT),
            (0, 1),
        )
        p_gn = tl.make_block_ptr(
            g + i_bh * s_k_h,
            (T * K,),
            (s_k_d,),
            ((i_t * BT + BT - 1) * K + i_k * BK,),
            (BK,),
            (0,),
        )

        tl.store(p_h, b_h.to(p_h.dtype.element_ty), boundary_check=(0, 1))
        # [BK, BT]
        b_k = tl.load(p_k, boundary_check=(0, 1))
        # [BT, BV]
        b_v = tl.load(p_v, boundary_check=(0, 1))
        # [BK, BT]
        b_g = tl.load(p_g, boundary_check=(0, 1))
        if i_t < NT - 1:
            # [BK,]
            b_gn = tl.load(p_gn, boundary_check=(0,))
        else:
            b_gn = tl.min(b_g, axis=1)
        b_h *= tl.exp(b_gn)[:, None]
        b_k = (b_k * tl.exp(b_gn[:, None] - b_g)).to(b_k.dtype)
        b_h += tl.dot(b_k, b_v, allow_tf32=False)

    if STORE_FINAL_STATE:
        p_h = tl.make_block_ptr(
            ht + i_bh * K * V, (K, V), (V, 1), (i_k * BK, i_v * BV), (BK, BV), (1, 0)
        )
        tl.store(p_h, b_h.to(p_h.dtype.element_ty), boundary_check=(0, 1))


@triton.jit
def chunk_gla_fwd_kernel_intra(
    q,
    k,
    g,
    A,
    s_k_h,
    s_k_t,
    s_k_d,
    scale,
    T: tl.constexpr,
    K: tl.constexpr,
    BT: tl.constexpr,
    BC: tl.constexpr,
    BK: tl.constexpr,
    NC: tl.constexpr,
):
    i_k, i_c, i_bh = tl.program_id(0), tl.program_id(1), tl.program_id(2)
    i_t, i_i, i_j = i_c // (NC * NC), (i_c % (NC * NC)) // NC, (i_c % (NC * NC)) % NC
    n_bh = tl.num_programs(2)

    if i_i > i_j:
        p_q = tl.make_block_ptr(
            q + i_bh * s_k_h,
            (T, K),
            (s_k_t, s_k_d),
            (i_t * BT + i_i * BC, i_k * BK),
            (BC, BK),
            (1, 0),
        )
        p_g = tl.make_block_ptr(
            g + i_bh * s_k_h,
            (T, K),
            (s_k_t, s_k_d),
            (i_t * BT + i_i * BC, i_k * BK),
            (BC, BK),
            (1, 0),
        )
        p_k = tl.make_block_ptr(
            k + i_bh * s_k_h,
            (K, T),
            (s_k_d, s_k_t),
            (i_k * BK, i_t * BT + i_j * BC),
            (BK, BC),
            (0, 1),
        )
        p_gk = tl.make_block_ptr(
            g + i_bh * s_k_h,
            (K, T),
            (s_k_d, s_k_t),
            (i_k * BK, i_t * BT + i_j * BC),
            (BK, BC),
            (0, 1),
        )
        p_gn = tl.make_block_ptr(
            g + i_bh * s_k_h,
            (T * K,),
            (s_k_d,),
            ((i_t * BT + i_i * BC) * K + i_k * BK,),
            (BK,),
            (0,),
        )
        p_A = tl.make_block_ptr(
            A + (i_k * n_bh + i_bh) * T * BT,
            (T, BT),
            (BT, 1),
            (i_t * BT + i_i * BC, i_j * BC),
            (BC, BC),
            (1, 0),
        )
        # [BK,]
        b_gn = tl.load(p_gn, boundary_check=(0,))
        # [BC, BK]
        b_q = tl.load(p_q, boundary_check=(0, 1))
        b_g = tl.load(p_g, boundary_check=(0, 1))
        b_qg = (b_q * tl.exp(b_g - b_gn[None, :]) * scale).to(b_q.dtype)
        # [BK, BC]
        b_k = tl.load(p_k, boundary_check=(0, 1))
        b_gk = tl.load(p_gk, boundary_check=(0, 1))
        b_kg = (b_k * tl.exp(b_gn[:, None] - b_gk)).to(b_k.dtype)
        # [BC, BC]
        b_A = tl.dot(b_qg, b_kg, allow_tf32=False)
        tl.store(p_A, b_A.to(A.dtype.element_ty), boundary_check=(0, 1))
    elif i_i == i_j:
        p_q = tl.make_block_ptr(
            q + i_bh * s_k_h,
            (T, K),
            (s_k_t, s_k_d),
            (i_t * BT + i_i * BC, i_k * BK),
            (BC, BK),
            (1, 0),
        )
        p_g = tl.make_block_ptr(
            g + i_bh * s_k_h,
            (T, K),
            (s_k_t, s_k_d),
            (i_t * BT + i_i * BC, i_k * BK),
            (BC, BK),
            (1, 0),
        )
        p_k = tl.make_block_ptr(
            k + i_bh * s_k_h,
            (T * K,),
            (s_k_d,),
            ((i_t * BT + i_j * BC) * K + i_k * BK,),
            (BK,),
            (0,),
        )
        p_gk = tl.make_block_ptr(
            g + i_bh * s_k_h,
            (T * K,),
            (s_k_d,),
            ((i_t * BT + i_j * BC) * K + i_k * BK,),
            (BK,),
            (0,),
        )
        # [BC, BK]
        b_q = tl.load(p_q, boundary_check=(0, 1))
        b_g = tl.load(p_g, boundary_check=(0, 1))

        o_i = tl.arange(0, BC)
        o_A = (
            (i_bh + i_k * n_bh) * T * BT
            + (i_t * BT + i_i * BC + tl.arange(0, BC)) * BT
            + i_j * BC
        )
        m_A = (i_t * BT + i_i * BC + tl.arange(0, BC)) < T
        for j in range(0, BC):
            # [BK,]
            b_k = tl.load(p_k, boundary_check=(0,)).to(tl.float32)
            b_gk = tl.load(p_gk, boundary_check=(0,)).to(tl.float32)
            # [BC,]
            b_A = tl.sum(b_q * b_k[None, :] * tl.exp(b_g - b_gk[None, :]) * scale, 1)
            b_A = tl.where(o_i >= j, b_A, 0.0)
            tl.store(A + o_A + j, b_A.to(b_q.dtype), mask=m_A)

            p_k = tl.advance(p_k, (K,))
            p_gk = tl.advance(p_gk, (K,))


@triton.jit
def chunk_gla_fwd_kernel_inter(
    q,
    v,
    g,
    h,
    o,
    A,
    s_k_h,
    s_k_t,
    s_k_d,
    s_v_h,
    s_v_t,
    s_v_d,
    s_h_h,
    s_h_t,
    s_h_d,
    scale,
    T: tl.constexpr,
    K: tl.constexpr,
    V: tl.constexpr,
    BT: tl.constexpr,
    BK: tl.constexpr,
    BV: tl.constexpr,
):
    i_v, i_t, i_bh = tl.program_id(0), tl.program_id(1), tl.program_id(2)

    b_o = tl.zeros([BT, BV], dtype=tl.float32)
    for i_k in range(tl.cdiv(K, BK)):
        p_q = tl.make_block_ptr(
            q + i_bh * s_k_h,
            (T, K),
            (s_k_t, s_k_d),
            (i_t * BT, i_k * BK),
            (BT, BK),
            (1, 0),
        )
        p_g = tl.make_block_ptr(
            g + i_bh * s_k_h,
            (T, K),
            (s_k_t, s_k_d),
            (i_t * BT, i_k * BK),
            (BT, BK),
            (1, 0),
        )
        p_h = tl.make_block_ptr(
            h + i_bh * s_h_h + i_t * K * V,
            (K, V),
            (s_h_t, s_h_d),
            (i_k * BK, i_v * BV),
            (BK, BV),
            (1, 0),
        )

        # [BT, BK]
        b_q = tl.load(p_q, boundary_check=(0, 1))
        b_q = (b_q * scale).to(b_q.dtype)
        # [BT, BK]
        b_g = tl.load(p_g, boundary_check=(0, 1))
        # [BT, BK]
        b_qg = (b_q * tl.exp(b_g)).to(b_q.dtype)
        # [BK, BV]
        b_h = tl.load(p_h, boundary_check=(0, 1))
        # works but dkw, owing to divine benevolence
        # [BT, BV]
        if i_k >= 0:
            b_o += tl.dot(b_qg, b_h, allow_tf32=False)
    p_v = tl.make_block_ptr(
        v + i_bh * s_v_h, (T, V), (s_v_t, s_v_d), (i_t * BT, i_v * BV), (BT, BV), (1, 0)
    )
    p_o = tl.make_block_ptr(
        o + i_bh * s_v_h, (T, V), (s_v_t, s_v_d), (i_t * BT, i_v * BV), (BT, BV), (1, 0)
    )
    p_A = tl.make_block_ptr(
        A + i_bh * T * BT, (T, BT), (BT, 1), (i_t * BT, 0), (BT, BT), (1, 0)
    )
    # [BT, BV]
    b_v = tl.load(p_v, boundary_check=(0, 1))
    # [BT, BT]
    b_A = tl.load(p_A, boundary_check=(0, 1))
    b_o += tl.dot(b_A, b_v, allow_tf32=False)
    tl.store(p_o, b_o.to(p_o.dtype.element_ty), boundary_check=(0, 1))


@triton.jit
def chunk_gla_bwd_kernel_dh(
    q,
    g,
    do,
    dh,
    s_k_h,
    s_k_t,
    s_k_d,
    s_v_h,
    s_v_t,
    s_v_d,
    s_h_h,
    s_h_t,
    s_h_d,
    scale,
    T: tl.constexpr,
    K: tl.constexpr,
    V: tl.constexpr,
    BT: tl.constexpr,
    BK: tl.constexpr,
    BV: tl.constexpr,
    NT: tl.constexpr,
):
    i_k, i_v, i_bh = tl.program_id(0), tl.program_id(1), tl.program_id(2)

    b_dh = tl.zeros([BK, BV], dtype=tl.float32)
    for i_t in range(NT - 1, -1, -1):
        p_q = tl.make_block_ptr(
            q + i_bh * s_k_h,
            (K, T),
            (s_k_d, s_k_t),
            (i_k * BK, i_t * BT),
            (BK, BT),
            (0, 1),
        )
        p_do = tl.make_block_ptr(
            do + i_bh * s_v_h,
            (T, V),
            (s_v_t, s_v_d),
            (i_t * BT, i_v * BV),
            (BT, BV),
            (1, 0),
        )
        p_dh = tl.make_block_ptr(
            dh + i_bh * s_h_h + i_t * K * V,
            (K, V),
            (s_h_t, s_h_d),
            (i_k * BK, i_v * BV),
            (BK, BV),
            (1, 0),
        )
        p_g = tl.make_block_ptr(
            g + i_bh * s_k_h,
            (K, T),
            (s_k_d, s_k_t),
            (i_k * BK, i_t * BT),
            (BK, BT),
            (0, 1),
        )
        p_gn = tl.make_block_ptr(
            g + i_bh * s_k_h,
            (T * K,),
            (s_k_d,),
            ((i_t * BT + BT - 1) * K + i_k * BK,),
            (BK,),
            (0,),
        )

        # [BK, BT]
        b_q = tl.load(p_q, boundary_check=(0, 1))
        b_q = (b_q * scale).to(b_q.dtype)
        # [BT, BV]
        b_do = tl.load(p_do, boundary_check=(0, 1))

        tl.store(p_dh, b_dh.to(p_dh.dtype.element_ty), boundary_check=(0, 1))

        # [BK,]
        b_gn = tl.load(p_gn, boundary_check=(0,))
        # [BK, BV]
        b_dh *= tl.exp(b_gn)[:, None]
        # [BK, BT]
        b_g = tl.load(p_g, boundary_check=(0, 1))
        b_q = (b_q * tl.exp(b_g)).to(b_q.dtype)

        # [BK, BV]
        b_dh += tl.dot(b_q, b_do, allow_tf32=False)


@triton.jit
def chunk_gla_bwd_kernel_inter(
    k,
    v,
    h,
    g,
    A,
    do,
    dh,
    dq,
    dk,
    dv,
    dA,
    s_k_h,
    s_k_t,
    s_k_d,
    s_v_h,
    s_v_t,
    s_v_d,
    s_h_h,
    s_h_t,
    s_h_d,
    scale,
    T: tl.constexpr,
    K: tl.constexpr,
    V: tl.constexpr,
    BT: tl.constexpr,
    BK: tl.constexpr,
    BV: tl.constexpr,
):
    i_k, i_t, i_bh = tl.program_id(0), tl.program_id(1), tl.program_id(2)
    n_bh = tl.num_programs(2)

    p_k = tl.make_block_ptr(
        k + i_bh * s_k_h, (T, K), (s_k_t, s_k_d), (i_t * BT, i_k * BK), (BT, BK), (1, 0)
    )
    p_gk = tl.make_block_ptr(
        g + i_bh * s_k_h, (T, K), (s_k_t, s_k_d), (i_t * BT, i_k * BK), (BT, BK), (1, 0)
    )
    p_gn = tl.make_block_ptr(
        g + i_bh * s_k_h,
        (T * K,),
        (s_k_d,),
        ((i_t * BT + BT - 1) * K + i_k * BK,),
        (BK,),
        (0,),
    )
    p_A = tl.make_block_ptr(
        A + i_bh * T * BT, (BT, T), (1, BT), (0, i_t * BT), (BT, BT), (0, 1)
    )

    # [BT, BK]
    b_k = tl.load(p_k, boundary_check=(0, 1))
    b_gk = tl.load(p_gk, boundary_check=(0, 1))
    b_gn = tl.exp(tl.load(p_gn, boundary_check=(0,))[None, :] - b_gk)
    b_k = (b_k * b_gn).to(b_k.dtype)
    # [BT, BT]
    b_A = tl.load(p_A, boundary_check=(0, 1))

    b_dq = tl.zeros([BT, BK], dtype=tl.float32)
    b_dk = tl.zeros([BT, BK], dtype=tl.float32)
    b_dA = tl.zeros([BT, BT], dtype=tl.float32)
    for i_v in range(tl.cdiv(V, BV)):
        p_v = tl.make_block_ptr(
            v + i_bh * s_v_h,
            (T, V),
            (s_v_t, s_v_d),
            (i_t * BT, i_v * BV),
            (BT, BV),
            (1, 0),
        )
        p_h = tl.make_block_ptr(
            h + i_bh * s_h_h + i_t * V * K,
            (V, K),
            (s_h_d, s_h_t),
            (i_v * BV, i_k * BK),
            (BV, BK),
            (0, 1),
        )
        p_do = tl.make_block_ptr(
            do + i_bh * s_v_h,
            (T, V),
            (s_v_t, s_v_d),
            (i_t * BT, i_v * BV),
            (BT, BV),
            (1, 0),
        )
        p_dh = tl.make_block_ptr(
            dh + i_bh * s_h_h + i_t * K * V,
            (K, V),
            (s_h_t, s_h_d),
            (i_k * BK, i_v * BV),
            (BK, BV),
            (1, 0),
        )
        p_dv = tl.make_block_ptr(
            dv + (i_k * n_bh + i_bh) * s_v_h,
            (T, V),
            (s_v_t, s_v_d),
            (i_t * BT, i_v * BV),
            (BT, BV),
            (1, 0),
        )

        # [BT, BV]
        b_v = tl.load(p_v, boundary_check=(0, 1))
        # [BV, BK]
        b_h = tl.load(p_h, boundary_check=(0, 1))
        # [BT, BV]
        b_do = tl.load(p_do, boundary_check=(0, 1))
        # [BK, BV]
        b_dh = tl.load(p_dh, boundary_check=(0, 1))

        # [BT, BV]
        b_dv = tl.dot(b_k, b_dh, allow_tf32=False)
        if i_k == 0:
            b_dv += tl.dot(b_A, b_do, allow_tf32=False)
        b_do = (b_do * scale).to(b_do.dtype)
        tl.store(p_dv, b_dv.to(p_dv.dtype.element_ty), boundary_check=(0, 1))
        # [BT, BT]
        b_dA += tl.dot(b_do, tl.trans(b_v), allow_tf32=False)
        # [BT, BK]
        b_dq += tl.dot(b_do, b_h, allow_tf32=False)
        # [BT, BK]
        b_dk += tl.dot(b_v, tl.trans(b_dh), allow_tf32=False)
    b_dq = b_dq * tl.exp(b_gk)
    b_dk = b_dk * b_gn

    p_dq = tl.make_block_ptr(
        dq + i_bh * s_k_h,
        (T, K),
        (s_k_t, s_k_d),
        (i_t * BT, i_k * BK),
        (BT, BK),
        (1, 0),
    )
    p_dk = tl.make_block_ptr(
        dk + i_bh * s_k_h,
        (T, K),
        (s_k_t, s_k_d),
        (i_t * BT, i_k * BK),
        (BT, BK),
        (1, 0),
    )
    p_dA = tl.make_block_ptr(
        dA + i_bh * T * BT,
        (
            T,
            BT,
        ),
        (BT, 1),
        (i_t * BT, 0),
        (BT, BT),
        (1, 0),
    )
    tl.store(p_dq, b_dq.to(p_dq.dtype.element_ty), boundary_check=(0, 1))
    tl.store(p_dk, b_dk.to(p_dk.dtype.element_ty), boundary_check=(0, 1))

    o_i = tl.arange(0, BT)
    m_s = o_i[:, None] >= o_i[None, :]
    # [BT, BT]
    b_dA = tl.where(m_s, b_dA, 0.0).to(b_k.dtype)
    if i_k == 0:
        tl.store(p_dA, b_dA.to(p_dA.dtype.element_ty), boundary_check=(0, 1))


@triton.jit
def chunk_gla_bwd_kernel_intra(
    q,
    k,
    g,
    dA,
    dq,
    dk,
    dg,
    s_k_h,
    s_k_t,
    s_k_d,
    T: tl.constexpr,
    K: tl.constexpr,
    BT: tl.constexpr,
    BC: tl.constexpr,
    BK: tl.constexpr,
    NC: tl.constexpr,
):
    i_k, i_c, i_bh = tl.program_id(0), tl.program_id(1), tl.program_id(2)
    i_t, i_i = i_c // NC, i_c % NC

    p_g = tl.make_block_ptr(
        g + i_bh * s_k_h,
        (T, K),
        (s_k_t, s_k_d),
        (i_t * BT + i_i * BC, i_k * BK),
        (BC, BK),
        (1, 0),
    )
    p_gn = tl.make_block_ptr(
        g + i_bh * s_k_h,
        (T * K,),
        (s_k_d,),
        ((i_t * BT + i_i * BC) * K + i_k * BK,),
        (BK,),
        (0,),
    )
    # [BK,]
    b_gn = tl.load(p_gn, boundary_check=(0,))
    # [BC, BK]
    b_g = tl.load(p_g, boundary_check=(0, 1))
    b_dq = tl.zeros([BC, BK], dtype=tl.float32)
    for i_j in range(0, i_i):
        p_k = tl.make_block_ptr(
            k + i_bh * s_k_h,
            (T, K),
            (s_k_t, s_k_d),
            (i_t * BT + i_j * BC, i_k * BK),
            (BC, BK),
            (1, 0),
        )
        p_gk = tl.make_block_ptr(
            g + i_bh * s_k_h,
            (T, K),
            (s_k_t, s_k_d),
            (i_t * BT + i_j * BC, i_k * BK),
            (BC, BK),
            (1, 0),
        )
        p_dA = tl.make_block_ptr(
            dA + i_bh * T * BT,
            (T, BT),
            (BT, 1),
            (i_t * BT + i_i * BC, i_j * BC),
            (BC, BC),
            (1, 0),
        )
        # [BC, BK]
        b_k = tl.load(p_k, boundary_check=(0, 1))
        b_gk = tl.load(p_gk, boundary_check=(0, 1))
        b_kg = (b_k * tl.exp(b_gn[None, :] - b_gk)).to(b_k.dtype)
        # [BC, BC]
        b_dA = tl.load(p_dA, boundary_check=(0, 1))
        # [BC, BK]
        b_dq += tl.dot(b_dA, b_kg, allow_tf32=False)
    b_dq *= tl.exp(b_g - b_gn[None, :])

    o_i = tl.arange(0, BC)
    o_dA = i_bh * T * BT + (i_t * BT + i_i * BC + tl.arange(0, BC)) * BT + i_i * BC
    m_dA = (i_t * BT + i_i * BC + tl.arange(0, BC)) < T
    for j in range(0, BC):
        p_kj = tl.make_block_ptr(
            k + i_bh * s_k_h,
            (T * K,),
            (1,),
            ((i_t * BT + i_i * BC + j) * K + i_k * BK,),
            (BK,),
            (0,),
        )
        p_gkj = tl.make_block_ptr(
            g + i_bh * s_k_h,
            (T * K,),
            (1,),
            ((i_t * BT + i_i * BC + j) * K + i_k * BK,),
            (BK,),
            (0,),
        )
        # [BC,]
        b_dA = tl.load(dA + o_dA + j, mask=m_dA, other=0)
        # [BK,]
        b_kj = tl.load(p_kj, boundary_check=(0,)).to(tl.float32)
        b_gkj = tl.load(p_gkj, boundary_check=(0,)).to(tl.float32)
        # [BC, BK]
        m_i = o_i[:, None] >= j
        # [BC, BK]
        b_dq += tl.where(
            m_i, b_dA[:, None] * b_kj[None, :] * tl.exp(b_g - b_gkj[None, :]), 0.0
        )
    p_dq = tl.make_block_ptr(
        dq + i_bh * s_k_h,
        (T, K),
        (s_k_t, s_k_d),
        (i_t * BT + i_i * BC, i_k * BK),
        (BC, BK),
        (1, 0),
    )

    b_dq = b_dq + tl.load(p_dq, boundary_check=(0, 1))
    tl.store(p_dq, b_dq.to(p_dq.dtype.element_ty), boundary_check=(0, 1))

    tl.debug_barrier()
    p_k = tl.make_block_ptr(
        k + i_bh * s_k_h,
        (T, K),
        (s_k_t, s_k_d),
        (i_t * BT + i_i * BC, i_k * BK),
        (BC, BK),
        (1, 0),
    )
    p_gk = tl.make_block_ptr(
        g + i_bh * s_k_h,
        (T, K),
        (s_k_t, s_k_d),
        (i_t * BT + i_i * BC, i_k * BK),
        (BC, BK),
        (1, 0),
    )
    p_gn = tl.make_block_ptr(
        g + i_bh * s_k_h,
        (T * K,),
        (s_k_d,),
        ((i_t * BT + i_i * BC + BC - 1) * K + i_k * BK,),
        (BK,),
        (0,),
    )
    # [BK,]
    b_gn = tl.load(p_gn, boundary_check=(0,))
    # [BC, BK]
    b_k = tl.load(p_k, boundary_check=(0, 1))
    b_gk = tl.load(p_gk, boundary_check=(0, 1))
    b_dk = tl.zeros([BC, BK], dtype=tl.float32)
    for i_j in range(i_i + 1, NC):
        p_q = tl.make_block_ptr(
            q + i_bh * s_k_h,
            (T, K),
            (s_k_t, s_k_d),
            (i_t * BT + i_j * BC, i_k * BK),
            (BC, BK),
            (1, 0),
        )
        p_g = tl.make_block_ptr(
            g + i_bh * s_k_h,
            (T, K),
            (s_k_t, s_k_d),
            (i_t * BT + i_j * BC, i_k * BK),
            (BC, BK),
            (1, 0),
        )
        p_dA = tl.make_block_ptr(
            dA + i_bh * T * BT,
            (T, BT),
            (BT, 1),
            (i_t * BT + i_j * BC, i_i * BC),
            (BC, BC),
            (1, 0),
        )
        # [BC, BK]
        b_q = tl.load(p_q, boundary_check=(0, 1))
        b_g = tl.load(p_g, boundary_check=(0, 1))
        b_qg = (b_q * tl.exp(b_g - b_gn[None, :])).to(b_q.dtype)
        # [BC, BC]
        b_dA = tl.load(p_dA, boundary_check=(0, 1))
        # [BC, BK]
        b_dk += tl.dot(tl.trans(b_dA), b_qg, allow_tf32=False)
    b_dk *= tl.exp(b_gn[None, :] - b_gk)

    o_dA = i_bh * T * BT + (i_t * BT + i_i * BC) * BT + i_i * BC + tl.arange(0, BC)
    for j in range(0, BC):
        p_qj = tl.make_block_ptr(
            q + i_bh * s_k_h,
            (T * K,),
            (1,),
            ((i_t * BT + i_i * BC + j) * K + i_k * BK,),
            (BK,),
            (0,),
        )
        p_gqj = tl.make_block_ptr(
            g + i_bh * s_k_h,
            (T * K,),
            (1,),
            ((i_t * BT + i_i * BC + j) * K + i_k * BK,),
            (BK,),
            (0,),
        )
        # [BC,]
        b_dA = tl.load(dA + o_dA + j * BT, mask=(i_t * BT + i_i * BC + j < T), other=0)
        # [BK,]
        b_qj = tl.load(p_qj, boundary_check=(0,)).to(tl.float32)
        b_gqj = tl.load(p_gqj, boundary_check=(0,)).to(tl.float32)
        # [BC, BK]
        m_i = o_i[:, None] <= j
        b_dk += tl.where(
            m_i, b_dA[:, None] * b_qj[None, :] * tl.exp(b_gqj[None, :] - b_gk), 0.0
        )

    p_q = tl.make_block_ptr(
        q + i_bh * s_k_h,
        (T, K),
        (s_k_t, s_k_d),
        (i_t * BT + i_i * BC, i_k * BK),
        (BC, BK),
        (1, 0),
    )
    p_dk = tl.make_block_ptr(
        dk + i_bh * s_k_h,
        (T, K),
        (s_k_t, s_k_d),
        (i_t * BT + i_i * BC, i_k * BK),
        (BC, BK),
        (1, 0),
    )
    p_dg = tl.make_block_ptr(
        dg + i_bh * s_k_h,
        (T, K),
        (s_k_t, s_k_d),
        (i_t * BT + i_i * BC, i_k * BK),
        (BC, BK),
        (1, 0),
    )

    b_q = tl.load(p_q, boundary_check=(0, 1))
    b_dk = b_dk + tl.load(p_dk, boundary_check=(0, 1))
    b_dg = b_q * b_dq - b_k * b_dk
    tl.store(p_dk, b_dk.to(p_dk.dtype.element_ty), boundary_check=(0, 1))
    tl.store(p_dg, b_dg.to(p_dg.dtype.element_ty), boundary_check=(0, 1))


class ChunkGLAFunction(torch.autograd.Function):
    @staticmethod
    @contiguous
    def forward(
        ctx, q, k, v, g, scale, initial_state, output_final_state, checkpoint_level
    ):
        B, H, T, K, V = *q.shape, v.shape[-1]
        BT, BC = 64, 16
        BK = min(64, triton.next_power_of_2(K))
        BV = min(64, triton.next_power_of_2(V))
        NT, NC = triton.cdiv(T, BT), triton.cdiv(BT, BC)
        NK = triton.cdiv(K, BK)
        NV = triton.cdiv(V, BV)
        num_warps = 4 if BK == 64 else 2
        num_stages = 1

        def fwd_inner(q, k, v, g, B, H, T, K, V, BT, BK, BV, NT, h0=None, ht=None):
            NK, NV = triton.cdiv(K, BK), triton.cdiv(V, BV)
            h = q.new_empty(B, H, NT * K, V)
            grid = (NV, NK, B * H)
            chunk_gla_fwd_kernel_h[grid](
                k,
                v,
                g,
                h,
                h0,
                ht,
                k.stride(1),
                k.stride(2),
                k.stride(3),
                v.stride(1),
                v.stride(2),
                v.stride(3),
                h.stride(1),
                h.stride(2),
                h.stride(3),
                T=T,
                K=K,
                V=V,
                BT=BT,
                BK=BK,
                BV=BV,
                NT=NT,
                USE_INITIAL_STATE=h0 is not None,
                STORE_FINAL_STATE=ht is not None,
                num_warps=num_warps,
                num_stages=num_stages,
            )
            return h

        final_state = None
        if output_final_state:
            final_state = q.new_empty(B, H, K, V, dtype=torch.float)

        g_org, g = g, torch.empty_like(g, dtype=torch.float)

        def grid(meta):
            return (triton.cdiv(meta["S"], meta["BS"]), NT, B * H)

        # keep cummulative normalizer in fp32
        # this kernel is equivalent to
        # g = g.view(B, H, NT, BT, -1).cumsum(-2).view(B, H, T, -1)
        chunk_gla_fwd_kernel_cum[grid](
            g_org, g, g.stride(1), g.stride(2), g.stride(3), T=T, S=K, BT=BT
        )
        h = fwd_inner(
            q=q,
            k=k,
            v=v,
            g=g,
            B=B,
            H=H,
            T=T,
            K=K,
            V=V,
            BT=BT,
            BK=BK,
            BV=BV,
            NT=NT,
            h0=initial_state if initial_state is not None else None,
            ht=final_state if final_state is not None else None,
        )
        A = q.new_zeros(NK, B, H, T, BT)
        grid = (NK, NT * NC * NC, B * H)
        chunk_gla_fwd_kernel_intra[grid](
            q,
            k,
            g,
            A,
            k.stride(1),
            k.stride(2),
            k.stride(3),
            scale,
            T=T,
            K=K,
            BT=BT,
            BC=BC,
            BK=BK,
            NC=NC,
            num_warps=num_warps,
            num_stages=num_stages,
        )
        A = A.sum(0, dtype=A.dtype)
        o = torch.empty_like(v)
        grid = (NV, NT, B * H)
        chunk_gla_fwd_kernel_inter[grid](
            q,
            v,
            g,
            h,
            o,
            A,
            k.stride(1),
            k.stride(2),
            k.stride(3),
            v.stride(1),
            v.stride(2),
            v.stride(3),
            h.stride(1),
            h.stride(2),
            h.stride(3),
            scale,
            T=T,
            K=K,
            V=V,
            BT=BT,
            BK=BK,
            BV=BV,
            num_warps=num_warps,
            num_stages=num_stages,
        )
        if checkpoint_level >= 1:
            del g
            g = g_org
        if checkpoint_level > 1:
            del h
            h, initial_state = None, None

        ctx.save_for_backward(q, k, v, g, h, initial_state, A)
        ctx.BT = BT
        ctx.scale = scale
        ctx.checkpoint_level = checkpoint_level
        return o, final_state

    @staticmethod
    @contiguous
    def backward(ctx, do, dht=None):
        q, k, v, g, h, initial_state, A = ctx.saved_tensors
        B, H, T, K, V = *q.shape, v.shape[-1]
        BT, BC = ctx.BT, 16
        BK = min(64, triton.next_power_of_2(K))
        BV = min(64, triton.next_power_of_2(V))
        NT, NC = triton.cdiv(T, BT), triton.cdiv(BT, BC)
        NK = triton.cdiv(K, BK)
        num_warps = 4 if BK == 64 else 2
        num_stages = 1

        def fwd_inner(q, k, v, g, B, H, T, K, V, BT, BK, BV, NT, h0=None, ht=None):
            NK, NV = triton.cdiv(K, BK), triton.cdiv(V, BV)
            h = q.new_empty(B, H, NT * K, V)
            grid = (NV, NK, B * H)
            chunk_gla_fwd_kernel_h[grid](
                k,
                v,
                g,
                h,
                h0,
                ht,
                k.stride(1),
                k.stride(2),
                k.stride(3),
                v.stride(1),
                v.stride(2),
                v.stride(3),
                h.stride(1),
                h.stride(2),
                h.stride(3),
                T=T,
                K=K,
                V=V,
                BT=BT,
                BK=BK,
                BV=BV,
                NT=NT,
                USE_INITIAL_STATE=h0 is not None,
                STORE_FINAL_STATE=ht is not None,
                num_warps=num_warps,
                num_stages=num_stages,
            )
            return h

        def bwd_inner(q, g, do, B, H, T, K, V, BT, BK, BV, NT, scale):
            NK, NV = triton.cdiv(K, BK), triton.cdiv(V, BV)
            dh = q.new_empty(B, H, NT * K, V)
            grid = (NK, NV, B * H)
            chunk_gla_bwd_kernel_dh[grid](
                q,
                g,
                do,
                dh,
                q.stride(1),
                q.stride(2),
                q.stride(3),
                do.stride(1),
                do.stride(2),
                do.stride(3),
                dh.stride(1),
                dh.stride(2),
                dh.stride(3),
                scale,
                T=T,
                K=K,
                V=V,
                BT=BT,
                BK=BK,
                BV=BV,
                NT=NT,
                num_warps=num_warps,
                num_stages=num_stages,
            )
            return dh

        if ctx.checkpoint_level >= 1:
            # save the original g and compute its fp32 cumsum during the backward pass for memory consideration
            g_org, g = g, torch.zeros_like(g, dtype=torch.float)

            def grid(meta):
                return (triton.cdiv(meta["S"], meta["BS"]), NT, B * H)

            # keep cummulative normalizer in fp32
            # this kernel is equivalent to
            # g = g.view(B, H, NT, BT, -1).cumsum(-2).view(B, H, T, -1)
            chunk_gla_fwd_kernel_cum[grid](
                g_org, g, g.stride(1), g.stride(2), g.stride(3), T=T, S=K, BT=BT
            )

        # rerun the forward pass to get h if checkpoint_level >= 1
        if ctx.checkpoint_level > 1:
            h = fwd_inner(
                q=q,
                k=k,
                v=v,
                g=g,
                B=B,
                H=H,
                T=T,
                K=K,
                V=V,
                BT=BT,
                BK=BK,
                BV=BV,
                NT=NT,
                h0=initial_state if initial_state is not None else None,
                ht=None,
            )

        scale = ctx.scale
        dh = bwd_inner(
            q, g, do, B=B, H=H, T=T, K=K, V=V, BT=BT, BK=BK, BV=BV, NT=NT, scale=scale
        )
        dq = torch.empty_like(q, dtype=torch.float)
        dk = torch.empty_like(k, dtype=torch.float)
        dg = torch.empty_like(k, dtype=torch.float)
        dv = v.new_empty(NK, *v.shape)
        dA = q.new_zeros(B, H, T, BT)
        grid = (NK, NT, B * H)
        chunk_gla_bwd_kernel_inter[grid](
            k,
            v,
            h,
            g,
            A,
            do,
            dh,
            dq,
            dk,
            dv,
            dA,
            k.stride(1),
            k.stride(2),
            k.stride(3),
            v.stride(1),
            v.stride(2),
            v.stride(3),
            h.stride(1),
            h.stride(2),
            h.stride(3),
            scale,
            T=T,
            K=K,
            V=V,
            BT=BT,
            BK=BK,
            BV=BV,
            num_warps=num_warps,
            num_stages=num_stages,
        )
        dv = dv.sum(0, dtype=dv.dtype)
        grid = (NK, NT * NC, B * H)
        chunk_gla_bwd_kernel_intra[grid](
            q,
            k,
            g,
            dA,
            dq,
            dk,
            dg,
            k.stride(1),
            k.stride(2),
            k.stride(3),
            T=T,
            K=K,
            BT=BT,
            BC=BC,
            BK=BK,
            NC=NC,
            num_warps=num_warps,
            num_stages=num_stages,
        )

        dq = dq.to(q.dtype)
        dk = dk.to(q.dtype)
        # reversed cumsum, equivalent to:
        #
        # def reversed_cumsum(x, dim=-1):
        #     c = x.cumsum(dim)
        #     return x + c.index_select(dim, x.new_tensor([c.shape[dim]-1], dtype=torch.long)) - c
        dg = chunk_reversed_cumsum_fwd(dg).to(k.dtype)
        return dq, dk, dv, dg, None, None, None, None


def chunk_gla(
    q: torch.Tensor,
    k: torch.Tensor,
    v: torch.Tensor,
    g: torch.Tensor,
    scale: int | None = None,
    initial_state: torch.Tensor = None,
    output_final_state: bool = False,
    checkpoint_level: int | None = 2,
) -> tuple[torch.Tensor, torch.Tensor]:
    r"""
    Args:
        q (torch.Tensor):
            queries of shape `(B, H, T, K)`
        k (torch.Tensor):
            keys of shape `(B, H, T, K)`
        v (torch.Tensor):
            values of shape `(B, H, T, V)`
        g (torch.Tensor):
            Forget gates of shape `(B, H, T, K)` applied to keys.
        scale (Optional[int]):
            Scale factor for the GLA attention scores.
            If not provided, it will default to `1 / sqrt(K)`. Default: `None`.
        initial_state (Optional[torch.Tensor]):
            Initial state of shape `(B, H, K, V)`. Default: `None`.
        output_final_state (Optional[bool]):
            Whether to output the final state of shape `(B, H, K, V)`. Default: `False`.
        checkpoint_level (Optional[int]):
            Checkpointing level; higher values will save more memories and do more recomputations during backward.
            Default: `0`:
            - Level `0`: no memory saved, no recomputation.
            - Level `1`: recompute the fp32 cumulative values during backward.
            - Level `2`: recompute the fp32 cumulative values and forward hidden states during backward.
    """
    assert checkpoint_level in [0, 1, 2]
    if scale is None:
        scale = q.shape[-1] ** -0.5
    if initial_state is not None:
        initial_state = initial_state.detach()
    o, final_state = ChunkGLAFunction.apply(
        q, k, v, g, scale, initial_state, output_final_state, checkpoint_level
    )
    return o, final_state
